@@ -42,6 +42,7 @@
 #include "avdtp.h"
 #include "sink.h"
 #include "a2dp.h"
+#include "sdpd.h"
 
 /* The duration that streams without users are allowed to stay in
  * STREAMING state. */
@@ -130,7 +131,7 @@ static void setup_unref(struct a2dp_setup *setup)
 		setup_free(setup);
 }
 
-static struct device *a2dp_get_dev(struct avdtp *session)
+static struct audio_device *a2dp_get_dev(struct avdtp *session)
 {
 	bdaddr_t addr;
 
@@ -241,13 +242,13 @@ static struct a2dp_setup *find_setup_by_session(struct avdtp *session)
 	return NULL;
 }
 
-static struct a2dp_setup *find_setup_by_dev(struct device *dev)
+static struct a2dp_setup *find_setup_by_dev(struct audio_device *dev)
 {
 	GSList *l;
 
 	for (l = setups; l != NULL; l = l->next) {
 		struct a2dp_setup *setup = l->data;
-		struct device *setup_dev = a2dp_get_dev(setup->session);
+		struct audio_device *setup_dev = a2dp_get_dev(setup->session);
 
 		if (setup_dev == dev)
 			return setup;
@@ -288,7 +289,7 @@ static gboolean sbc_setconf_ind(struct avdtp *session,
 				uint8_t *category, void *user_data)
 {
 	struct a2dp_sep *a2dp_sep = user_data;
-	struct device *dev;
+	struct audio_device *dev;
 	struct avdtp_service_capability *cap;
 	struct avdtp_media_codec_capability *codec_cap;
 	struct sbc_codec_cap *sbc_cap;
@@ -394,7 +395,7 @@ static gboolean mpeg_setconf_ind(struct avdtp *session,
 				uint8_t *category, void *user_data)
 {
 	struct a2dp_sep *a2dp_sep = user_data;
-	struct device *dev;
+	struct audio_device *dev;
 
 	if (a2dp_sep->type == AVDTP_SEP_TYPE_SINK)
 		debug("Sink %p: Set_Configuration_Ind", sep);
@@ -472,7 +473,7 @@ static void setconf_cfm(struct avdtp *session, struct avdtp_local_sep *sep,
 {
 	struct a2dp_sep *a2dp_sep = user_data;
 	struct a2dp_setup *setup;
-	struct device *dev;
+	struct audio_device *dev;
 	int ret;
 
 	if (a2dp_sep->type == AVDTP_SEP_TYPE_SINK)
@@ -728,15 +729,29 @@ static gboolean close_ind(struct avdtp *session, struct avdtp_local_sep *sep,
 	return TRUE;
 }
 
-static gboolean reconfigure(gpointer data)
+static gboolean a2dp_reconfigure(gpointer data)
 {
 	struct a2dp_setup *setup = data;
 	struct avdtp_local_sep *lsep;
 	struct avdtp_remote_sep *rsep;
+	struct avdtp_service_capability *cap;
+	struct avdtp_media_codec_capability *codec_cap = NULL;
+	GSList *l;
 	int posix_err;
 
+	for (l = setup->client_caps; l != NULL; l = l->next) {
+		cap = l->data;
+
+		if (cap->category != AVDTP_MEDIA_CODEC)
+			continue;
+
+		codec_cap = (void *) cap->data;
+		break;
+	}
+
 	posix_err = avdtp_get_seps(setup->session, AVDTP_SEP_TYPE_SINK,
-					AVDTP_MEDIA_TYPE_AUDIO, A2DP_CODEC_SBC,
+					codec_cap->media_type,
+					codec_cap->media_codec_type,
 					&lsep, &rsep);
 	if (posix_err < 0) {
 		error("No matching ACP and INT SEPs found");
@@ -784,7 +799,7 @@ static void close_cfm(struct avdtp *session, struct avdtp_local_sep *sep,
 	}
 
 	if (setup->reconfigure)
-		g_timeout_add(RECONFIGURE_TIMEOUT, reconfigure, setup);
+		g_timeout_add(RECONFIGURE_TIMEOUT, a2dp_reconfigure, setup);
 }
 
 static gboolean abort_ind(struct avdtp *session, struct avdtp_local_sep *sep,
@@ -901,31 +916,32 @@ static struct avdtp_sep_ind mpeg_ind = {
 	.reconfigure		= reconf_ind
 };
 
-static int a2dp_source_record(sdp_buf_t *buf)
+static sdp_record_t *a2dp_source_record()
 {
 	sdp_list_t *svclass_id, *pfseq, *apseq, *root;
 	uuid_t root_uuid, l2cap, avdtp, a2src;
 	sdp_profile_desc_t profile[1];
 	sdp_list_t *aproto, *proto[2];
-	sdp_record_t record;
+	sdp_record_t *record;
 	sdp_data_t *psm, *version, *features;
 	uint16_t lp = AVDTP_UUID, ver = 0x0100, feat = 0x000F;
-	int ret = 0;
 
-	memset(&record, 0, sizeof(sdp_record_t));
+	record = sdp_record_alloc();
+	if (!record)
+		return NULL;
 
 	sdp_uuid16_create(&root_uuid, PUBLIC_BROWSE_GROUP);
 	root = sdp_list_append(0, &root_uuid);
-	sdp_set_browse_groups(&record, root);
+	sdp_set_browse_groups(record, root);
 
 	sdp_uuid16_create(&a2src, AUDIO_SOURCE_SVCLASS_ID);
 	svclass_id = sdp_list_append(0, &a2src);
-	sdp_set_service_classes(&record, svclass_id);
+	sdp_set_service_classes(record, svclass_id);
 
 	sdp_uuid16_create(&profile[0].uuid, ADVANCED_AUDIO_PROFILE_ID);
 	profile[0].version = 0x0100;
 	pfseq = sdp_list_append(0, &profile[0]);
-	sdp_set_profile_descs(&record, pfseq);
+	sdp_set_profile_descs(record, pfseq);
 
 	sdp_uuid16_create(&l2cap, L2CAP_UUID);
 	proto[0] = sdp_list_append(0, &l2cap);
@@ -940,17 +956,12 @@ static int a2dp_source_record(sdp_buf_t *buf)
 	apseq = sdp_list_append(apseq, proto[1]);
 
 	aproto = sdp_list_append(0, apseq);
-	sdp_set_access_protos(&record, aproto);
+	sdp_set_access_protos(record, aproto);
 
 	features = sdp_data_alloc(SDP_UINT16, &feat);
-	sdp_attr_add(&record, SDP_ATTR_SUPPORTED_FEATURES, features);
+	sdp_attr_add(record, SDP_ATTR_SUPPORTED_FEATURES, features);
 
-	sdp_set_info_attr(&record, "Audio Source", 0, 0);
-
-	if (sdp_gen_record_pdu(&record, buf) < 0)
-		ret = -1;
-	else
-		ret = 0;
+	sdp_set_info_attr(record, "Audio Source", 0, 0);
 
 	free(psm);
 	free(version);
@@ -961,15 +972,13 @@ static int a2dp_source_record(sdp_buf_t *buf)
 	sdp_list_free(aproto, 0);
 	sdp_list_free(root, 0);
 	sdp_list_free(svclass_id, 0);
-	sdp_list_free(record.attrlist, (sdp_free_func_t) sdp_data_free);
-	sdp_list_free(record.pattern, free);
 
-	return ret;
+	return record;
 }
 
-static int a2dp_sink_record(sdp_buf_t *buf)
+static sdp_record_t *a2dp_sink_record()
 {
-	return -1;
+	return NULL;
 }
 
 static struct a2dp_sep *a2dp_add_sep(DBusConnection *conn, uint8_t type,
@@ -977,9 +986,9 @@ static struct a2dp_sep *a2dp_add_sep(DBusConnection *conn, uint8_t type,
 {
 	struct a2dp_sep *sep;
 	GSList **l;
-	int (*create_record)(sdp_buf_t *buf);
+	sdp_record_t *(*create_record)(void);
 	uint32_t *record_id;
-	sdp_buf_t buf;
+	sdp_record_t *record;
 	struct avdtp_sep_ind *ind;
 
 	sep = g_new0(struct a2dp_sep, 1);
@@ -1008,22 +1017,22 @@ static struct a2dp_sep *a2dp_add_sep(DBusConnection *conn, uint8_t type,
 	if (*record_id != 0)
 		goto add;
 
-	memset(&buf, 0, sizeof(buf));
-	if (create_record(&buf) < 0) {
+	record = create_record();
+	if (!record) {
 		error("Unable to allocate new service record");
 		avdtp_unregister_sep(sep->sep);
 		g_free(sep);
 		return NULL;
 	}
 
-	*record_id = add_service_record(conn, &buf);
-	free(buf.data);
-	if (!*record_id) {
-		error("Unable to register A2DP service record");
+	if (add_record_to_server(BDADDR_ANY, record) < 0) {
+		error("Unable to register A2DP service record");\
+		sdp_record_free(record);
 		avdtp_unregister_sep(sep->sep);
 		g_free(sep);
 		return NULL;
 	}
+	*record_id = record->handle;
 
 add:
 	*l = g_slist_append(*l, sep);
@@ -1100,7 +1109,7 @@ int a2dp_init(DBusConnection *conn, GKeyFile *config)
 proceed:
 	connection = dbus_connection_ref(conn);
 
-	avdtp_init();
+	avdtp_init(config);
 
 	if (source) {
 		for (i = 0; i < sbc_srcs; i++)
@@ -1142,19 +1151,19 @@ void a2dp_exit()
 	sources = NULL;
 
 	if (source_record_id) {
-		remove_service_record(connection, source_record_id);
+		remove_record_from_server(source_record_id);
 		source_record_id = 0;
 	}
 
 	if (sink_record_id) {
-		remove_service_record(connection, sink_record_id);
+		remove_record_from_server(sink_record_id);
 		sink_record_id = 0;
 	}
 
 	dbus_connection_unref(connection);
 }
 
-gboolean a2dp_source_cancel(struct device *dev, unsigned int id)
+gboolean a2dp_source_cancel(struct audio_device *dev, unsigned int id)
 {
 	struct a2dp_setup_cb *cb_data;
 	struct a2dp_setup *setup;
@@ -1193,7 +1202,7 @@ unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
 	struct a2dp_setup_cb *cb_data;
 	GSList *l;
 	struct a2dp_setup *setup;
-	struct a2dp_sep *sep = NULL;
+	struct a2dp_sep *sep = NULL, *tmp;
 	struct avdtp_local_sep *lsep;
 	struct avdtp_remote_sep *rsep;
 	struct avdtp_service_capability *cap;
@@ -1214,7 +1223,7 @@ unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
 		return 0;
 
 	for (l = sources; l != NULL; l = l->next) {
-		struct a2dp_sep *tmp = l->data;
+		tmp = l->data;
 
 		if (tmp->locked)
 			continue;
@@ -1255,6 +1264,22 @@ unsigned int a2dp_source_config(struct avdtp *session, a2dp_config_cb_t cb,
 
 	switch (avdtp_sep_get_state(sep->sep)) {
 	case AVDTP_STATE_IDLE:
+		for (l = sources; l != NULL; l = l->next) {
+			tmp = l->data;
+
+			if (avdtp_has_stream(session, tmp->stream))
+				break;
+		}
+
+		if (l != NULL) {
+			setup->reconfigure = TRUE;
+			if (avdtp_close(session, tmp->stream) < 0) {
+				error("avdtp_close failed");
+				goto failed;
+			}
+			break;
+		}
+
 		if (avdtp_get_seps(session, AVDTP_SEP_TYPE_SINK,
 				codec_cap->media_type,
 				codec_cap->media_codec_type,
