@@ -421,7 +421,11 @@ static int bluetooth_prepare(snd_pcm_ioplug_t *io)
 	}
 
 	/* wake up any client polling at us */
-	return write(data->pipefd[1], &c, 1);
+	err = write(data->pipefd[1], &c, 1);
+	if (err < 0)
+		return err;
+
+	return 0;
 }
 
 static int bluetooth_hsp_hw_params(snd_pcm_ioplug_t *io,
@@ -503,10 +507,16 @@ static uint8_t default_bitpool(uint8_t freq, uint8_t mode)
 	}
 }
 
-static int select_sbc_params(sbc_capabilities_t *cap, unsigned int rate,
-				unsigned int channels)
+static int bluetooth_a2dp_init(struct bluetooth_data *data,
+				snd_pcm_hw_params_t *params)
 {
-	unsigned int max_bitpool, min_bitpool;
+	struct bluetooth_alsa_config *cfg = &data->alsa_config;
+	sbc_capabilities_t *cap = &data->a2dp.sbc_capabilities;
+	unsigned int max_bitpool, min_bitpool, rate, channels;
+	int dir;
+
+	snd_pcm_hw_params_get_rate(params, &rate, &dir);
+	snd_pcm_hw_params_get_channels(params, &channels);
 
 	switch (rate) {
 	case 48000:
@@ -526,7 +536,9 @@ static int select_sbc_params(sbc_capabilities_t *cap, unsigned int rate,
 		return -1;
 	}
 
-	if (channels == 2) {
+	if (cfg->has_channel_mode)
+		cap->channel_mode = cfg->channel_mode;
+	else if (channels == 2) {
 		if (cap->channel_mode & BT_A2DP_CHANNEL_MODE_JOINT_STEREO)
 			cap->channel_mode = BT_A2DP_CHANNEL_MODE_JOINT_STEREO;
 		else if (cap->channel_mode & BT_A2DP_CHANNEL_MODE_STEREO)
@@ -543,7 +555,9 @@ static int select_sbc_params(sbc_capabilities_t *cap, unsigned int rate,
 		return -1;
 	}
 
-	if (cap->block_length & BT_A2DP_BLOCK_LENGTH_16)
+	if (cfg->has_block_length)
+		cap->block_length = cfg->block_length;
+	else if (cap->block_length & BT_A2DP_BLOCK_LENGTH_16)
 		cap->block_length = BT_A2DP_BLOCK_LENGTH_16;
 	else if (cap->block_length & BT_A2DP_BLOCK_LENGTH_12)
 		cap->block_length = BT_A2DP_BLOCK_LENGTH_12;
@@ -556,6 +570,8 @@ static int select_sbc_params(sbc_capabilities_t *cap, unsigned int rate,
 		return -1;
 	}
 
+	if (cfg->has_subbands)
+		cap->subbands = cfg->subbands;
 	if (cap->subbands & BT_A2DP_SUBBANDS_8)
 		cap->subbands = BT_A2DP_SUBBANDS_8;
 	else if (cap->subbands & BT_A2DP_SUBBANDS_4)
@@ -565,14 +581,21 @@ static int select_sbc_params(sbc_capabilities_t *cap, unsigned int rate,
 		return -1;
 	}
 
+	if (cfg->has_allocation_method)
+		cap->allocation_method = cfg->allocation_method;
 	if (cap->allocation_method & BT_A2DP_ALLOCATION_LOUDNESS)
 		cap->allocation_method = BT_A2DP_ALLOCATION_LOUDNESS;
 	else if (cap->allocation_method & BT_A2DP_ALLOCATION_SNR)
 		cap->allocation_method = BT_A2DP_ALLOCATION_SNR;
 
-	min_bitpool = MAX(MIN_BITPOOL, cap->min_bitpool);
-	max_bitpool = MIN(default_bitpool(cap->frequency, cap->channel_mode),
-							cap->max_bitpool);
+	if (cfg->has_bitpool)
+		min_bitpool = max_bitpool = cfg->bitpool;
+	else {
+		min_bitpool = MAX(MIN_BITPOOL, cap->min_bitpool);
+		max_bitpool = MIN(default_bitpool(cap->frequency,
+					cap->channel_mode),
+					cap->max_bitpool);
+	}
 
 	cap->min_bitpool = min_bitpool;
 	cap->max_bitpool = max_bitpool;
@@ -580,6 +603,72 @@ static int select_sbc_params(sbc_capabilities_t *cap, unsigned int rate,
 	return 0;
 }
 
+static void bluetooth_a2dp_setup(struct bluetooth_a2dp *a2dp)
+{
+	sbc_capabilities_t active_capabilities = a2dp->sbc_capabilities;
+
+	if (a2dp->sbc_initialized)
+		sbc_reinit(&a2dp->sbc, 0);
+	else
+		sbc_init(&a2dp->sbc, 0);
+	a2dp->sbc_initialized = 1;
+
+	if (active_capabilities.frequency & BT_SBC_SAMPLING_FREQ_16000)
+		a2dp->sbc.frequency = SBC_FREQ_16000;
+
+	if (active_capabilities.frequency & BT_SBC_SAMPLING_FREQ_32000)
+		a2dp->sbc.frequency = SBC_FREQ_32000;
+
+	if (active_capabilities.frequency & BT_SBC_SAMPLING_FREQ_44100)
+		a2dp->sbc.frequency = SBC_FREQ_44100;
+
+	if (active_capabilities.frequency & BT_SBC_SAMPLING_FREQ_48000)
+		a2dp->sbc.frequency = SBC_FREQ_48000;
+
+	if (active_capabilities.channel_mode & BT_A2DP_CHANNEL_MODE_MONO)
+		a2dp->sbc.mode = SBC_MODE_MONO;
+
+	if (active_capabilities.channel_mode & BT_A2DP_CHANNEL_MODE_DUAL_CHANNEL)
+		a2dp->sbc.mode = SBC_MODE_DUAL_CHANNEL;
+
+	if (active_capabilities.channel_mode & BT_A2DP_CHANNEL_MODE_STEREO)
+		a2dp->sbc.mode = SBC_MODE_STEREO;
+
+	if (active_capabilities.channel_mode & BT_A2DP_CHANNEL_MODE_JOINT_STEREO)
+		a2dp->sbc.mode = SBC_MODE_JOINT_STEREO;
+
+	a2dp->sbc.allocation = active_capabilities.allocation_method
+				== BT_A2DP_ALLOCATION_SNR ? SBC_AM_SNR
+				: SBC_AM_LOUDNESS;
+
+	switch (active_capabilities.subbands) {
+	case BT_A2DP_SUBBANDS_4:
+		a2dp->sbc.subbands = SBC_SB_4;
+		break;
+	case BT_A2DP_SUBBANDS_8:
+		a2dp->sbc.subbands = SBC_SB_8;
+		break;
+	}
+
+	switch (active_capabilities.block_length) {
+	case BT_A2DP_BLOCK_LENGTH_4:
+		a2dp->sbc.blocks = SBC_BLK_4;
+		break;
+	case BT_A2DP_BLOCK_LENGTH_8:
+		a2dp->sbc.blocks = SBC_BLK_8;
+		break;
+	case BT_A2DP_BLOCK_LENGTH_12:
+		a2dp->sbc.blocks = SBC_BLK_12;
+		break;
+	case BT_A2DP_BLOCK_LENGTH_16:
+		a2dp->sbc.blocks = SBC_BLK_16;
+		break;
+	}
+
+	a2dp->sbc.bitpool = active_capabilities.max_bitpool;
+	a2dp->codesize = sbc_get_codesize(&a2dp->sbc);
+	a2dp->count = sizeof(struct rtp_header) + sizeof(struct rtp_payload);
+}
 
 static int bluetooth_a2dp_hw_params(snd_pcm_ioplug_t *io,
 					snd_pcm_hw_params_t *params)
@@ -590,29 +679,20 @@ static int bluetooth_a2dp_hw_params(snd_pcm_ioplug_t *io,
 	bt_audio_rsp_msg_header_t *rsp_hdr = (void*) buf;
 	struct bt_setconfiguration_req *setconf_req = (void*) buf;
 	struct bt_setconfiguration_rsp *setconf_rsp = (void*) buf;
-	unsigned int rate, channels;
-	int err, dir;
-	sbc_capabilities_t active_capabilities;
+	int err;
 
 	DBG("Preparing with io->period_size=%lu io->buffer_size=%lu",
 					io->period_size, io->buffer_size);
 
-	/* FIXME: this needs to be really implemented (take into account
-	real asoundrc settings + ALSA hw settings ) once server side sends us
-	more than one possible configuration */
-	snd_pcm_hw_params_get_rate(params, &rate, &dir);
-	snd_pcm_hw_params_get_channels(params, &channels);
-	err = select_sbc_params(&a2dp->sbc_capabilities, rate, channels);
+	err = bluetooth_a2dp_init(data, params);
 	if (err < 0)
 		return err;
 
-	active_capabilities = a2dp->sbc_capabilities;
-
 	memset(setconf_req, 0, BT_AUDIO_IPC_PACKET_SIZE);
 	setconf_req->h.msg_type = BT_SETCONFIGURATION_REQ;
 	strncpy(setconf_req->device, data->alsa_config.device, 18);
 	setconf_req->transport = BT_CAPABILITIES_TRANSPORT_A2DP;
-	setconf_req->sbc_capabilities = active_capabilities;
+	setconf_req->sbc_capabilities = a2dp->sbc_capabilities;
 	setconf_req->access_mode = (io->stream == SND_PCM_STREAM_PLAYBACK ?
 			BT_CAPABILITIES_ACCESS_MODE_WRITE :
 			BT_CAPABILITIES_ACCESS_MODE_READ);
@@ -637,67 +717,7 @@ static int bluetooth_a2dp_hw_params(snd_pcm_ioplug_t *io,
 	data->link_mtu = setconf_rsp->link_mtu;
 
 	/* Setup SBC encoder now we agree on parameters */
-	if (a2dp->sbc_initialized)
-		sbc_reinit(&a2dp->sbc, 0);
-	else
-		sbc_init(&a2dp->sbc, 0);
-	a2dp->sbc_initialized = 1;
-
-	if (active_capabilities.frequency & BT_SBC_SAMPLING_FREQ_16000)
-		a2dp->sbc.rate = 16000;
-
-	if (active_capabilities.frequency & BT_SBC_SAMPLING_FREQ_32000)
-		a2dp->sbc.rate = 32000;
-
-	if (active_capabilities.frequency & BT_SBC_SAMPLING_FREQ_44100)
-		a2dp->sbc.rate = 44100;
-
-	if (active_capabilities.frequency & BT_SBC_SAMPLING_FREQ_48000)
-		a2dp->sbc.rate = 48000;
-
-	if (active_capabilities.channel_mode & BT_A2DP_CHANNEL_MODE_MONO)
-		a2dp->sbc.channels = 1;
-	else
-		a2dp->sbc.channels = 2;
-
-	if (active_capabilities.channel_mode &
-			(BT_A2DP_CHANNEL_MODE_MONO ||
-			BT_A2DP_CHANNEL_MODE_JOINT_STEREO))
-		a2dp->sbc.joint = 1;
-	else
-		a2dp->sbc.joint = 0;
-
-	a2dp->sbc.allocation = active_capabilities.allocation_method
-					== BT_A2DP_ALLOCATION_SNR ? 0x01 : 0x00;
-
-	switch (active_capabilities.subbands) {
-	case BT_A2DP_SUBBANDS_4:
-		a2dp->sbc.subbands = 4;
-		break;
-	case BT_A2DP_SUBBANDS_8:
-		a2dp->sbc.subbands = 8;
-		break;
-	}
-
-	switch (active_capabilities.block_length) {
-	case BT_A2DP_BLOCK_LENGTH_4:
-		a2dp->sbc.blocks = 4;
-		break;
-	case BT_A2DP_BLOCK_LENGTH_8:
-		a2dp->sbc.blocks = 8;
-		break;
-	case BT_A2DP_BLOCK_LENGTH_12:
-		a2dp->sbc.blocks = 12;
-		break;
-	case BT_A2DP_BLOCK_LENGTH_16:
-		a2dp->sbc.blocks = 16;
-		break;
-	}
-
-	a2dp->sbc.bitpool = active_capabilities.max_bitpool;
-	a2dp->codesize = a2dp->sbc.subbands * a2dp->sbc.blocks *
-						a2dp->sbc.channels * 2;
-	a2dp->count = sizeof(struct rtp_header) + sizeof(struct rtp_payload);
+	bluetooth_a2dp_setup(a2dp);
 
 	DBG("\tallocation=%u\n\tsubbands=%u\n\tblocks=%u\n\tbitpool=%u\n",
 		a2dp->sbc.allocation, a2dp->sbc.subbands, a2dp->sbc.blocks,
@@ -734,6 +754,11 @@ static int bluetooth_poll_revents(snd_pcm_ioplug_t *io ATTRIBUTE_UNUSED,
 	return 0;
 }
 
+static int bluetooth_playback_poll_descriptors_count(snd_pcm_ioplug_t *io)
+{
+	return 2;
+}
+
 static int bluetooth_playback_poll_descriptors(snd_pcm_ioplug_t *io,
 					struct pollfd *pfd, unsigned int space)
 {
@@ -743,14 +768,17 @@ static int bluetooth_playback_poll_descriptors(snd_pcm_ioplug_t *io,
 
 	assert(data->pipefd[0] >= 0);
 
-	if (space < 1)
+	if (space < 2)
 		return 0;
 
 	pfd[0].fd = data->pipefd[0];
 	pfd[0].events = POLLIN;
 	pfd[0].revents = 0;
+	pfd[1].fd = data->stream.fd;
+	pfd[1].events = POLLERR | POLLHUP | POLLNVAL;
+	pfd[1].revents = 0;
 
-	return 1;
+	return 2;
 }
 
 static int bluetooth_playback_poll_revents(snd_pcm_ioplug_t *io,
@@ -763,14 +791,19 @@ static int bluetooth_playback_poll_revents(snd_pcm_ioplug_t *io,
 	DBG("");
 
 	assert(pfds);
-	assert(nfds == 1);
+	assert(nfds == 2);
 	assert(revents);
 	assert(pfds[0].fd >= 0);
+	assert(pfds[1].fd >= 0);
 
 	if (io->state != SND_PCM_STATE_PREPARED)
 		ret = read(pfds[0].fd, buf, 1);
 
-	*revents = (pfds[0].revents & ~POLLIN) | POLLOUT;
+	if (pfds[1].revents & (POLLERR | POLLHUP | POLLNVAL))
+		io->state = SND_PCM_STATE_DISCONNECTED;
+
+	revents[0] = (pfds[0].revents & ~POLLIN) | POLLOUT;
+	revents[1] = (pfds[1].revents & ~POLLIN);
 
 	return 0;
 }
@@ -1071,6 +1104,7 @@ static snd_pcm_ioplug_callback_t bluetooth_hsp_playback = {
 	.hw_params		= bluetooth_hsp_hw_params,
 	.prepare		= bluetooth_prepare,
 	.transfer		= bluetooth_hsp_write,
+	.poll_descriptors_count	= bluetooth_playback_poll_descriptors_count,
 	.poll_descriptors	= bluetooth_playback_poll_descriptors,
 	.poll_revents		= bluetooth_playback_poll_revents,
 	.delay			= bluetooth_playback_delay,
@@ -1096,6 +1130,7 @@ static snd_pcm_ioplug_callback_t bluetooth_a2dp_playback = {
 	.hw_params		= bluetooth_a2dp_hw_params,
 	.prepare		= bluetooth_prepare,
 	.transfer		= bluetooth_a2dp_write,
+	.poll_descriptors_count	= bluetooth_playback_poll_descriptors_count,
 	.poll_descriptors	= bluetooth_playback_poll_descriptors,
 	.poll_revents		= bluetooth_playback_poll_revents,
 	.delay			= bluetooth_playback_delay,
@@ -1172,6 +1207,7 @@ static int bluetooth_a2dp_hw_constraint(snd_pcm_ioplug_t *io)
 {
 	struct bluetooth_data *data = io->private_data;
 	struct bluetooth_a2dp *a2dp = &data->a2dp;
+	struct bluetooth_alsa_config *cfg = &data->alsa_config;
 	snd_pcm_access_t access_list[] = {
 		SND_PCM_ACCESS_RW_INTERLEAVED,
 		/* Mmap access is really useless fo this driver, but we
@@ -1204,12 +1240,17 @@ static int bluetooth_a2dp_hw_constraint(snd_pcm_ioplug_t *io)
 		return err;
 
 	/* supported channels */
-	if (a2dp->sbc_capabilities.channel_mode & BT_A2DP_CHANNEL_MODE_MONO)
+	if (cfg->has_channel_mode)
+		a2dp->sbc_capabilities.channel_mode = cfg->channel_mode;
+
+	if (a2dp->sbc_capabilities.channel_mode &
+			BT_A2DP_CHANNEL_MODE_MONO)
 		min_channels = 1;
 	else
 		min_channels = 2;
 
-	if (a2dp->sbc_capabilities.channel_mode & (~BT_A2DP_CHANNEL_MODE_MONO))
+	if (a2dp->sbc_capabilities.channel_mode &
+			(~BT_A2DP_CHANNEL_MODE_MONO))
 		max_channels = 2;
 	else
 		max_channels = 1;
@@ -1234,28 +1275,33 @@ static int bluetooth_a2dp_hw_constraint(snd_pcm_ioplug_t *io)
 
 	/* supported rates */
 	rate_count = 0;
-	if (a2dp->sbc_capabilities.frequency &
-			BT_SBC_SAMPLING_FREQ_16000) {
-		rate_list[rate_count] = 16000;
+	if (cfg->has_rate) {
+		rate_list[rate_count] = cfg->rate;
 		rate_count++;
-	}
+	} else {
+		if (a2dp->sbc_capabilities.frequency &
+				BT_SBC_SAMPLING_FREQ_16000) {
+			rate_list[rate_count] = 16000;
+			rate_count++;
+		}
 
-	if (a2dp->sbc_capabilities.frequency &
-			BT_SBC_SAMPLING_FREQ_32000) {
-		rate_list[rate_count] = 32000;
-		rate_count++;
-	}
+		if (a2dp->sbc_capabilities.frequency &
+				BT_SBC_SAMPLING_FREQ_32000) {
+			rate_list[rate_count] = 32000;
+			rate_count++;
+		}
 
-	if (a2dp->sbc_capabilities.frequency &
-			BT_SBC_SAMPLING_FREQ_44100) {
-		rate_list[rate_count] = 44100;
-		rate_count++;
-	}
+		if (a2dp->sbc_capabilities.frequency &
+				BT_SBC_SAMPLING_FREQ_44100) {
+			rate_list[rate_count] = 44100;
+			rate_count++;
+		}
 
-	if (a2dp->sbc_capabilities.frequency &
-			BT_SBC_SAMPLING_FREQ_48000) {
-		rate_list[rate_count] = 48000;
-		rate_count++;
+		if (a2dp->sbc_capabilities.frequency &
+				BT_SBC_SAMPLING_FREQ_48000) {
+			rate_list[rate_count] = 48000;
+			rate_count++;
+		}
 	}
 
 	err = snd_pcm_ioplug_set_param_list(io, SND_PCM_IOPLUG_HW_RATE,
@@ -1348,10 +1394,7 @@ static int bluetooth_parse_config(snd_config_t *conf,
 				return -EINVAL;
 			}
 
-			if (strcmp(value, "auto") == 0) {
-				bt_config->channel_mode = BT_A2DP_CHANNEL_MODE_AUTO;
-				bt_config->has_channel_mode = 1;
-			} else if (strcmp(value, "mono") == 0) {
+			if (strcmp(value, "mono") == 0) {
 				bt_config->channel_mode = BT_A2DP_CHANNEL_MODE_MONO;
 				bt_config->has_channel_mode = 1;
 			} else if (strcmp(value, "dual") == 0) {
@@ -1373,10 +1416,7 @@ static int bluetooth_parse_config(snd_config_t *conf,
 				return -EINVAL;
 			}
 
-			if (strcmp(value, "auto") == 0) {
-				bt_config->allocation_method = BT_A2DP_ALLOCATION_AUTO;
-				bt_config->has_allocation_method = 1;
-			} else if (strcmp(value, "loudness") == 0) {
+			if (strcmp(value, "loudness") == 0) {
 				bt_config->allocation_method = BT_A2DP_ALLOCATION_LOUDNESS;
 				bt_config->has_allocation_method = 1;
 			} else if (strcmp(value, "snr") == 0) {
