@@ -11,6 +11,7 @@
 
 struct _GModule {
 	void *handle;
+	gchar *file_name;
 };
 
 static const char *dl_error_string = NULL;
@@ -33,6 +34,8 @@ GModule *g_module_open(const gchar *file_name, GModuleFlags flags)
 		return NULL;
 	}
 
+	module->file_name = g_strdup(file_name);
+
 	return module;
 }
 
@@ -60,6 +63,7 @@ gboolean g_module_close(GModule *module)
 		return FALSE;
 	}
 
+	g_free(module->file_name);
 	g_free(module);
 
 	return TRUE;
@@ -74,3 +78,11 @@ const gchar *g_module_error(void)
 
 	return str;
 }
+
+const gchar *g_module_name(GModule *module)
+{
+	if (module == NULL)
+		return NULL;
+
+	return module->file_name;
+}
