@@ -25,8 +25,11 @@
 #define DATABASE_INTERFACE "org.bluez.Database"
 
 dbus_bool_t database_init(DBusConnection *conn, const char *path);
+void database_cleanup(DBusConnection *conn, const char *path);
 
-DBusHandlerResult database_message(DBusConnection *conn,
-						DBusMessage *msg, void *data);
-
-void set_sdp_server_enable(void);
+int add_xml_record(DBusConnection *conn, const char *sender, bdaddr_t *src,
+				const char *record, dbus_uint32_t *handle);
+DBusMessage *update_xml_record(DBusConnection *conn,
+				DBusMessage *msg, bdaddr_t *src);
+int remove_record(DBusConnection *conn, const char *sender,
+						dbus_uint32_t handle);
