@@ -46,11 +46,9 @@
 #include <netinet/in.h>
 
 #include <glib.h>
-
 #include <dbus/dbus.h>
+#include <gdbus.h>
 
-#include "dbus.h"
-#include "dbus-helper.h"
 #include "hcid.h"
 #include "textfile.h"
 #include "adapter.h"
@@ -60,11 +58,14 @@
 #include "error.h"
 #include "dbus-sdp.h"
 #include "sdp-xml.h"
+#include "glib-helper.h"
 
 #define SESSION_TIMEOUT 2000
 #define DEFAULT_XML_BUF_SIZE	1024
 
 struct transaction_context {
+	char		*src;
+	char		*dst;
 	DBusConnection	*conn;
 	DBusMessage	*rq;
 	sdp_session_t	*session;
@@ -77,21 +78,14 @@ struct transaction_context {
 typedef int connect_cb_t(struct transaction_context *t);
 
 struct pending_connect {
-	DBusConnection *conn;
-	DBusMessage *rq;
-
-	char *dst;
-	sdp_session_t *session;
-	connect_cb_t *conn_cb;
+	DBusConnection	*conn;
+	DBusMessage	*rq;
+	char		*src;
+	char		*dst;
+	sdp_session_t	*session;
+	connect_cb_t	*conn_cb;
 };
 
-/* FIXME:  move to a common file */
-typedef struct {
-	char            *name;
-	uint16_t        class;
-	char            *info_name;
-} sdp_service_t;
-
 struct cached_session {
 	sdp_session_t *session;
 	guint timeout_id;
@@ -100,6 +94,29 @@ struct cached_session {
 
 static GSList *cached_sessions = NULL;
 
+static inline DBusMessage *invalid_args(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
+			"Invalid arguments in method call");
+}
+
+static inline DBusMessage *in_progress(DBusMessage *msg, const char *str)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".InProgress", str);
+}
+
+static inline DBusMessage *adapter_not_ready(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".NotReady",
+			"Adapter is not ready");
+}
+
+static inline DBusMessage *failed_strerror(DBusMessage *msg, int err)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+			strerror(err));
+}
+
 static gboolean session_timeout(gpointer user_data)
 {
 	struct cached_session *s = user_data;
@@ -222,7 +239,7 @@ static sdp_session_t *get_sdp_session(bdaddr_t *src, bdaddr_t *dst)
 	return session;
 }
 
-static void append_and_grow_string(void *data, const char *str)
+void append_and_grow_string(void *data, const char *str)
 {
 	sdp_buf_t *buff = data;
 	int len;
@@ -257,43 +274,12 @@ static void append_and_grow_string(void *data, const char *str)
 	buff->data_size += len;
 }
 
-/* FIXME:  move to a common file */
-sdp_service_t sdp_service[] = {
-	{ "vcp",	VIDEO_CONF_SVCLASS_ID,		"Video Conference"	},
-	{ "map",	0,				NULL			},
-	{ "pbap",	PBAP_SVCLASS_ID,		"Phone Book Access"	},
-	{ "sap",	SAP_SVCLASS_ID,			"SIM Access"		},
-	{ "ftp",	OBEX_FILETRANS_SVCLASS_ID,	"OBEX File Transfer"	},
-	{ "bpp",	BASIC_PRINTING_SVCLASS_ID,	"Printing"		},
-	{ "bip",	IMAGING_SVCLASS_ID,		"Imaging"		},
-	{ "synch",	IRMC_SYNC_SVCLASS_ID,		"Synchronization"	},
-	{ "dun",	DIALUP_NET_SVCLASS_ID,		"Dial-Up Networking"	},
-	{ "opp",	OBEX_OBJPUSH_SVCLASS_ID,	"OBEX Object Push"	},
-	{ "fax",	FAX_SVCLASS_ID,			"Fax"			},
-	{ "spp",	SERIAL_PORT_SVCLASS_ID,		"Serial Port"		},
-	{ "hsp",	HEADSET_SVCLASS_ID,		"Headset"		},
-	{ "hfp",	HANDSFREE_SVCLASS_ID,		"Handsfree"		},
-	{ NULL }
-};
-
-/* FIXME:  move to a common file */
-uint16_t sdp_str2svclass(const char *str)
-{
-	sdp_service_t *s;
-
-	for (s = sdp_service; s->name; s++) {
-		if (strcasecmp(s->name, str) == 0)
-			return s->class;
-	}
-
-	return 0;
-}
-
 /* list of remote and local service records */
 static GSList *pending_connects  = NULL;
 
-static struct pending_connect *pending_connect_new(DBusConnection *conn, DBusMessage *msg,
-							const char *dst, connect_cb_t *cb)
+static struct pending_connect *pending_connect_new(DBusConnection *conn,
+					DBusMessage *msg, const char *src,
+					const char *dst, connect_cb_t *cb)
 {
 	struct pending_connect *c;
 
@@ -301,10 +287,8 @@ static struct pending_connect *pending_connect_new(DBusConnection *conn, DBusMes
 		return NULL;
 
 	c = g_new0(struct pending_connect, 1);
-
-	if (dst)
-		c->dst = g_strdup(dst);
-
+	c->src = g_strdup(src);
+	c->dst = g_strdup(dst);
 	c->conn = dbus_connection_ref(conn);
 	c->rq = dbus_message_ref(msg);
 	c->conn_cb = cb;
@@ -317,6 +301,7 @@ static void pending_connect_free(struct pending_connect *c)
 	if (!c)
 		return;
 
+	g_free(c->src);
 	g_free(c->dst);
 
 	if (c->rq)
@@ -341,21 +326,6 @@ static struct pending_connect *find_pending_connect(const char *dst)
 	return NULL;
 }
 
-static const char *get_address_from_message(DBusConnection *conn, DBusMessage *msg)
-{
-	struct adapter *adapter;
-	const char *path;
-
-	path = dbus_message_get_path(msg);
-	if (!path)
-		return NULL;
-
-	if (dbus_connection_get_object_user_data(conn, path, (void *) &adapter) == FALSE)
-		return NULL;
-
-	return adapter->address;
-}
-
 static int sdp_store_record(const char *src, const char *dst, uint32_t handle, uint8_t *buf, size_t size)
 {
 	char filename[PATH_MAX + 1], key[28], *value;
@@ -386,6 +356,9 @@ static void transaction_context_free(void *udata, gboolean cache)
 	if (!ctxt)
 		return;
 
+	g_free(ctxt->src);
+	g_free(ctxt->dst);
+
 	if (ctxt->conn)
 		dbus_connection_unref(ctxt->conn);
 
@@ -451,7 +424,6 @@ static void remote_svc_rec_completed_cb(uint8_t type, uint16_t err,
 	sdp_record_t *rec;
 	DBusMessage *reply;
 	DBusMessageIter iter, array_iter;
-	const char *src, *dst;
 	int scanned;
 
 	if (!ctxt)
@@ -483,24 +455,18 @@ static void remote_svc_rec_completed_cb(uint8_t type, uint16_t err,
 		goto failed;
 	}
 
-	dbus_message_get_args(ctxt->rq, NULL,
-			DBUS_TYPE_STRING, &dst,
-			DBUS_TYPE_INVALID);
-
-	src = get_address_from_message(ctxt->conn, ctxt->rq);
-
 	reply = dbus_message_new_method_return(ctxt->rq);
 	dbus_message_iter_init_append(reply, &iter);
 	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
 			DBUS_TYPE_BYTE_AS_STRING, &array_iter);
 
-	rec = sdp_extract_pdu(rsp, &scanned);
+	rec = sdp_extract_pdu_safe(rsp, size, &scanned);
 	if (rec == NULL || size != scanned) {
 		error("Invalid service record!");
 		goto done;
 	}
 
-	sdp_store_record(src, dst, rec->handle, rsp, size);
+	sdp_store_record(ctxt->src, ctxt->dst, rec->handle, rsp, size);
 
 	sdp_record_free(rec);
 
@@ -509,7 +475,8 @@ static void remote_svc_rec_completed_cb(uint8_t type, uint16_t err,
 
 done:
 	dbus_message_iter_close_container(&iter, &array_iter);
-	send_message_and_unref(ctxt->conn, reply);
+	dbus_connection_send(ctxt->conn, reply, NULL);
+	dbus_message_unref(reply);
 
 failed:
 	transaction_context_free(ctxt, TRUE);
@@ -522,7 +489,6 @@ static void remote_svc_rec_completed_xml_cb(uint8_t type, uint16_t err,
 	struct transaction_context *ctxt = udata;
 	sdp_record_t *rec;
 	DBusMessage *reply;
-	const char *src, *dst;
 	int scanned;
 	sdp_buf_t result;
 
@@ -555,21 +521,15 @@ static void remote_svc_rec_completed_xml_cb(uint8_t type, uint16_t err,
 		goto failed;
 	}
 
-	dbus_message_get_args(ctxt->rq, NULL,
-			DBUS_TYPE_STRING, &dst,
-			DBUS_TYPE_INVALID);
-
-	src = get_address_from_message(ctxt->conn, ctxt->rq);
-
 	reply = dbus_message_new_method_return(ctxt->rq);
 
-	rec = sdp_extract_pdu(rsp, &scanned);
+	rec = sdp_extract_pdu_safe(rsp, size, &scanned);
 	if (rec == NULL || size != scanned) {
 		error("Invalid service record!");
 		goto done;
 	}
 
-	sdp_store_record(src, dst, rec->handle, rsp, size);
+	sdp_store_record(ctxt->src, ctxt->dst, rec->handle, rsp, size);
 
 	memset(&result, 0, sizeof(sdp_buf_t));
 
@@ -585,7 +545,8 @@ static void remote_svc_rec_completed_xml_cb(uint8_t type, uint16_t err,
 		free(result.data);
 	}
 done:
-	send_message_and_unref(ctxt->conn, reply);
+	dbus_connection_send(ctxt->conn, reply, NULL);
+	dbus_message_unref(reply);
 
 failed:
 	transaction_context_free(ctxt, TRUE);
@@ -659,7 +620,8 @@ static void remote_svc_handles_completed_cb(uint8_t type, uint16_t err,
 
 done:
 	dbus_message_iter_close_container(&iter, &array_iter);
-	send_message_and_unref(ctxt->conn, reply);
+	dbus_connection_send(ctxt->conn, reply, NULL);
+	dbus_message_unref(reply);
 
 failed:
 	transaction_context_free(ctxt, TRUE);
@@ -724,12 +686,12 @@ static void remote_svc_identifiers_completed_cb(uint8_t type, uint16_t err,
 			uint8_t *rsp, size_t size, void *udata)
 {
 	struct transaction_context *ctxt = udata;
-	const char *src, *dst, *puuid;
+	const char *puuid;
 	const char *devid_uuid = "00001200-0000-1000-8000-00805f9b34fb";
 	char **identifiers;
 	DBusMessage *reply;
 	GSList *l = NULL;
-	int scanned, extracted = 0, len = 0, recsize = 0;
+	int scanned, extracted = 0, len = 0, recsize = 0, bytesleft = size;
 	uint8_t dtd = 0;
 
 	if (!ctxt)
@@ -761,23 +723,19 @@ static void remote_svc_identifiers_completed_cb(uint8_t type, uint16_t err,
 		goto failed;
 	}
 
-	src = get_address_from_message(ctxt->conn, ctxt->rq);
-	dbus_message_get_args(ctxt->rq, NULL,
-			DBUS_TYPE_STRING, &dst,
-			DBUS_TYPE_INVALID);
-
-	scanned = sdp_extract_seqtype(rsp, &dtd, &len);
+	scanned = sdp_extract_seqtype_safe(rsp, bytesleft, &dtd, &len);
 	rsp += scanned;
-	for (; extracted < len; rsp += recsize, extracted += recsize) {
+	bytesleft -= scanned;
+	for (; extracted < len; rsp += recsize, extracted += recsize, bytesleft -= recsize) {
 		sdp_record_t *rec;
 		sdp_data_t *d;
 
 		recsize = 0;
-		rec = sdp_extract_pdu(rsp, &recsize);
+		rec = sdp_extract_pdu_safe(rsp, bytesleft, &recsize);
 		if (!rec)
 			break;
 
-		sdp_store_record(src, dst, rec->handle, rsp, recsize);
+		sdp_store_record(ctxt->src, ctxt->dst, rec->handle, rsp, recsize);
 
 		d = sdp_data_get(rec, SDP_ATTR_SVCLASS_ID_LIST);
 		if (!d) {
@@ -823,14 +781,16 @@ static void remote_svc_identifiers_completed_cb(uint8_t type, uint16_t err,
 			DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
 			&identifiers, len,
 			DBUS_TYPE_INVALID);
-	send_message_and_unref(ctxt->conn, reply);
+
+	dbus_connection_send(ctxt->conn, reply, NULL);
+	dbus_message_unref(reply);
 
 	if (len)
-		dbus_connection_emit_signal(ctxt->conn,
+		g_dbus_emit_signal(ctxt->conn,
 				dbus_message_get_path(ctxt->rq),
 				ADAPTER_INTERFACE,
 				"RemoteIdentifiersUpdated",
-				DBUS_TYPE_STRING, &dst,
+				DBUS_TYPE_STRING, &ctxt->dst,
 				DBUS_TYPE_ARRAY, DBUS_TYPE_STRING,
 				&identifiers, len,
 				DBUS_TYPE_INVALID);
@@ -864,7 +824,8 @@ static gboolean sdp_client_connect_cb(GIOChannel *chan,
 	}
 
 	ctxt = g_new0(struct transaction_context, 1);
-
+	ctxt->src = g_strdup(c->src);
+	ctxt->dst = g_strdup(c->dst);
 	ctxt->conn = dbus_connection_ref(c->conn);
 	ctxt->rq = dbus_message_ref(c->rq);
 	ctxt->session = c->session;
@@ -902,7 +863,7 @@ done:
 
 static struct pending_connect *connect_request(DBusConnection *conn,
 					DBusMessage *msg,
-					uint16_t dev_id,
+					const char *src,
 					const char *dst,
 					connect_cb_t *cb, int *err)
 {
@@ -910,16 +871,15 @@ static struct pending_connect *connect_request(DBusConnection *conn,
 	bdaddr_t srcba, dstba;
 	GIOChannel *chan;
 
-	c = pending_connect_new(conn, msg, dst, cb);
+	c = pending_connect_new(conn, msg, src, dst, cb);
 	if (!c) {
 		if (err)
 			*err = ENOMEM;
 		return NULL;
 	}
 
-	hci_devba(dev_id, &srcba);
+	str2ba(src, &srcba);
 	str2ba(dst, &dstba);
-
 	c->session = get_sdp_session(&srcba, &dstba);
 	if (!c->session) {
 		if (err)
@@ -999,7 +959,7 @@ static int remote_svc_rec_conn_xml_cb(struct transaction_context *ctxt)
 	return 0;
 }
 
-DBusHandlerResult get_remote_svc_rec(DBusConnection *conn, DBusMessage *msg,
+DBusMessage *get_remote_svc_rec(DBusConnection *conn, DBusMessage *msg,
 				void *data, sdp_format_t format)
 {
 	struct adapter *adapter = data;
@@ -1009,28 +969,28 @@ DBusHandlerResult get_remote_svc_rec(DBusConnection *conn, DBusMessage *msg,
 	connect_cb_t *cb;
 
 	if (!adapter->up)
-		return error_not_ready(conn, msg);
+		return adapter_not_ready(msg);
 
 	if (!dbus_message_get_args(msg, NULL,
 			DBUS_TYPE_STRING, &dst,
 			DBUS_TYPE_UINT32, &handle,
 			DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg, NULL);
+		return invalid_args(msg);
 
 	if (find_pending_connect(dst))
-		return error_service_search_in_progress(conn, msg);
+		return in_progress(msg, "Service search in progress");
 
 	cb = remote_svc_rec_conn_cb;
 	if (format == SDP_FORMAT_XML)
 		cb = remote_svc_rec_conn_xml_cb;
 
-	if (!connect_request(conn, msg, adapter->dev_id,
+	if (!connect_request(conn, msg, adapter->address,
 				dst, cb, &err)) {
 		error("Search request failed: %s (%d)", strerror(err), err);
-		return error_failed_errno(conn, msg, err);
+		return failed_strerror(msg, err);
 	}
 
-	return DBUS_HANDLER_RESULT_HANDLED;
+	return NULL;
 }
 
 static int remote_svc_handles_conn_cb(struct transaction_context *ctxt)
@@ -1047,7 +1007,7 @@ static int remote_svc_handles_conn_cb(struct transaction_context *ctxt)
 			DBUS_TYPE_INVALID);
 
 	if (strlen(svc) > 0)
-		str2uuid(&ctxt->uuid, svc);
+		bt_string2uuid(&ctxt->uuid, svc);
 	else
 		sdp_uuid16_create(&ctxt->uuid, PUBLIC_BROWSE_GROUP);
 
@@ -1074,7 +1034,8 @@ static int remote_svc_identifiers_conn_cb(struct transaction_context *ctxt)
 	return service_search_attr(ctxt, PUBLIC_BROWSE_GROUP);
 }
 
-DBusHandlerResult get_remote_svc_handles(DBusConnection *conn, DBusMessage *msg, void *data)
+DBusMessage *get_remote_svc_handles(DBusConnection *conn,
+				DBusMessage *msg, void *data)
 {
 	struct adapter *adapter = data;
 	const char *dst, *svc;
@@ -1082,77 +1043,73 @@ DBusHandlerResult get_remote_svc_handles(DBusConnection *conn, DBusMessage *msg,
 	uuid_t uuid;
 
 	if (!adapter->up)
-		return error_not_ready(conn, msg);
+		return adapter_not_ready(msg);
 
 	if (!dbus_message_get_args(msg, NULL,
 			DBUS_TYPE_STRING, &dst,
 			DBUS_TYPE_STRING, &svc,
 			DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg, NULL);
+		return invalid_args(msg);
 
 	if (strlen(svc) > 0) {
 		/* Check if it is a service name string */
-		if (str2uuid(&uuid, svc) < 0) {
+		if (bt_string2uuid(&uuid, svc) < 0) {
 			error("Invalid service class name");
-			return error_invalid_arguments(conn, msg, NULL);
+			return invalid_args(msg);
 		}
 	}
 
 	if (find_pending_connect(dst))
-		return error_service_search_in_progress(conn, msg);
+		return in_progress(msg, "Service search in progress");
 
-	if (!connect_request(conn, msg, adapter->dev_id,
+	if (!connect_request(conn, msg, adapter->address,
 				dst, remote_svc_handles_conn_cb, &err)) {
 		error("Search request failed: %s (%d)", strerror(err), err);
-		return error_failed_errno(conn, msg, err);
+		return failed_strerror(msg, err);
 	}
 
-	return DBUS_HANDLER_RESULT_HANDLED;
+	return NULL;
 }
 
-DBusHandlerResult get_remote_svc_identifiers(DBusConnection *conn, DBusMessage *msg, void *data)
+DBusMessage *get_remote_svc_identifiers(DBusConnection *conn,
+					DBusMessage *msg, void *data)
 {
 	struct adapter *adapter = data;
 	const char *dst;
 	int err;
 
 	if (!adapter->up)
-		return error_not_ready(conn, msg);
+		return adapter_not_ready(msg);
 
 	if (!dbus_message_get_args(msg, NULL,
 			DBUS_TYPE_STRING, &dst,
 			DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg, NULL);
+		return invalid_args(msg);
 
 	if (find_pending_connect(dst))
-		return error_service_search_in_progress(conn, msg);
+		return in_progress(msg, "Service search in progress");
 
-	if (!connect_request(conn, msg, adapter->dev_id,
+	if (!connect_request(conn, msg, adapter->address,
 				dst, remote_svc_identifiers_conn_cb, &err)) {
 		error("Search request failed: %s (%d)", strerror(err), err);
-		return error_failed_errno(conn, msg, err);
+		return failed_strerror(msg, err);
 	}
 
-	return DBUS_HANDLER_RESULT_HANDLED;
+	return NULL;
 }
 
-DBusHandlerResult finish_remote_svc_transact(DBusConnection *conn,
-						DBusMessage *msg, void *data)
+DBusMessage *finish_remote_svc_transact(DBusConnection *conn,
+					DBusMessage *msg, void *data)
 {
 	struct cached_session *s;
 	const char *address;
 	struct adapter *adapter = data;
-	DBusMessage *reply;
 	bdaddr_t sba, dba;
 
 	if (!dbus_message_get_args(msg, NULL,
 			DBUS_TYPE_STRING, &address,
 			DBUS_TYPE_INVALID))
-		return error_invalid_arguments(conn, msg, NULL);
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return invalid_args(msg);
 
 	str2ba(adapter->address, &sba);
 	str2ba(address, &dba);
@@ -1164,5 +1121,5 @@ DBusHandlerResult finish_remote_svc_transact(DBusConnection *conn,
 		g_free(s);
 	}
 
-	return send_message_and_unref(conn, reply);
+	return dbus_message_new_method_return(msg);
 }
