@@ -22,42 +22,26 @@
  *
  */
 
-#define START_REPLY_TIMEOUT	5000
-
 struct service {
-	char *filename;
 	char *object_path;
-
-	DBusMessage *action;	/* Either Start or Stop method call */
-
-	guint startup_timer;
-	guint shutdown_timer;
-
-	/* These are set when the service is running */
-	GPid pid;		/* Process id */
-	char *bus_name;		/* D-Bus unique name */
-
-	/* Information parsed from the service file */
-	char *name;
-	char *descr;
 	char *ident;
-	gboolean autostart;
-
-	/* Services without a *.service file */
-	gboolean external;
+	char *name;
 };
 
 void release_services(DBusConnection *conn);
 
 void append_available_services(DBusMessageIter *iter);
 
-struct service *search_service(DBusConnection *conn, const char *pattern);
+struct service *search_service(const char *pattern);
 
-int service_start(struct service *service, DBusConnection *conn);
-
-int init_services(const char *path);
-
-int service_register(DBusConnection *conn, const char *bus_name, const char *ident,
-				const char *name, const char *description);
+struct service *search_service_by_uuid(const char *uuid);
 
 int service_unregister(DBusConnection *conn, struct service *service);
+
+int register_service(const char *ident, const char **uuids);
+void unregister_service(const char *ident);
+
+typedef void (*service_auth_cb) (DBusError *derr, void *user_data);
+int service_req_auth(const bdaddr_t *src, const bdaddr_t *dst,
+		const char *uuid, service_auth_cb cb, void *user_data);
+int service_cancel_auth(const bdaddr_t *src, const bdaddr_t *dst);
