@@ -48,16 +48,15 @@
 
 #include <dbus/dbus.h>
 
-#include "notify.h"
-
 #include "hcid.h"
 #include "sdpd.h"
-#include "server.h"
 #include "adapter.h"
 #include "dbus-common.h"
 #include "dbus-service.h"
 #include "dbus-database.h"
 #include "dbus-hci.h"
+#include "device.h"
+#include "agent.h"
 
 struct hcid_opts hcid;
 struct device_opts default_device;
@@ -65,6 +64,23 @@ struct device_opts *parser_device;
 static struct device_list *device_list = NULL;
 static int child_pipe[2];
 
+static GKeyFile *load_config(const char *file)
+{
+	GError *err = NULL;
+	GKeyFile *keyfile;
+
+	keyfile = g_key_file_new();
+
+	if (!g_key_file_load_from_file(keyfile, file, 0, &err)) {
+		error("Parsing %s failed: %s", file, err->message);
+		g_error_free(err);
+		g_key_file_free(keyfile);
+		return NULL;
+	}
+
+	return keyfile;
+}
+
 static inline void init_device_defaults(struct device_opts *device_opts)
 {
 	memset(device_opts, 0, sizeof(*device_opts));
@@ -277,6 +293,9 @@ void update_service_classes(const bdaddr_t *bdaddr, uint8_t value)
 		if (hci_test_bit(HCI_RAW, &di.flags))
 			continue;
 
+		if (!hci_test_bit(HCI_UP, &di.flags))
+			continue;
+
 		if (get_device_class(di.dev_id, cls) < 0)
 			continue;
 
@@ -759,7 +778,7 @@ static void usage(void)
 {
 	printf("hcid - HCI daemon ver %s\n", VERSION);
 	printf("Usage: \n");
-	printf("\thcid [-n] [-d] [-s] [-m mtu] [-f config file]\n");
+	printf("\thcid [-n] [-d] [-m mtu] [-f config file]\n");
 }
 
 int main(int argc, char *argv[])
@@ -769,7 +788,8 @@ int main(int argc, char *argv[])
 	struct sigaction sa;
 	GIOChannel *ctl_io, *child_io;
 	uint16_t mtu = 0;
-	int opt, daemonize = 1, debug = 0, sdp = 0, experimental = 0;
+	int opt, daemonize = 1, debug = 0, sdp = 1, experimental = 0;
+	GKeyFile *config;
 
 	/* Default HCId settings */
 	memset(&hcid, 0, sizeof(hcid));
@@ -872,6 +892,8 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
+	config = load_config(CONFIGDIR "/main.conf");
+
 	if (read_config(hcid.config_file) < 0)
 		error("Config load failed");
 
@@ -889,17 +911,27 @@ int main(int argc, char *argv[])
 
 	init_adapters();
 
+	agent_init();
+
+	if (experimental)
+		hcid_dbus_set_experimental();
+
 	if (hcid_dbus_init() < 0) {
 		error("Unable to get on D-Bus");
 		exit(1);
 	}
 
-	if (experimental)
-		hcid_dbus_set_experimental();
+	start_sdp_server(mtu, hcid.deviceid, SDP_SERVER_COMPAT);
+	set_service_classes_callback(update_service_classes);
+
+	/* Loading plugins has to be done after D-Bus has been setup since
+	 * the plugins might wanna expose some paths on the bus. However the
+	 * best order of how to init various subsystems of the Bluetooth
+	 * daemon needs to be re-worked. */
+	plugin_init(config);
 
 	init_security_data();
 
-	/* Create event loop */
 	event_loop = g_main_loop_new(NULL, FALSE);
 
 	ctl_io = g_io_channel_unix_new(hcid.sock);
@@ -912,36 +944,25 @@ int main(int argc, char *argv[])
 	/* Initialize already connected devices */
 	init_all_devices(hcid.sock);
 
-	if (sdp) {
-		set_sdp_server_enable();
-		start_sdp_server(mtu, hcid.deviceid, SDP_SERVER_COMPAT);
-		set_service_classes_callback(update_service_classes);
-	}
-
-	notify_init();
-
-	init_local_server();
-
-	init_services(CONFIGDIR);
-
-	/* Start event processor */
 	g_main_loop_run(event_loop);
 
-	if (sdp)
-		stop_sdp_server();
+	hcid_dbus_unregister();
+
+	plugin_cleanup();
+
+	stop_sdp_server();
 
 	free_device_opts();
 
+	agent_exit();
+
 	hcid_dbus_exit();
 
-	notify_close();
-
-	shutdown_local_server();
-
-	cleanup_sdp_session();
-
 	g_main_loop_unref(event_loop);
 
+	if (config)
+		g_key_file_free(config);
+
 	info("Exit");
 
 	stop_logging();
