@@ -45,7 +45,7 @@
 #include "textfile.h"
 #include "hcid.h"
 
-static inline int create_filename(char *buf, size_t size, bdaddr_t *bdaddr, const char *name)
+static inline int create_filename(char *buf, size_t size, const bdaddr_t *bdaddr, const char *name)
 {
 	char addr[18];
 
@@ -433,7 +433,7 @@ int write_lastused_info(bdaddr_t *local, bdaddr_t *peer, struct tm *tm)
 	return textfile_put(filename, addr, str);
 }
 
-int write_link_key(bdaddr_t *local, bdaddr_t *peer, unsigned char *key, int type, int length)
+int write_link_key(bdaddr_t *local, bdaddr_t *peer, unsigned char *key, uint8_t type, int length)
 {
 	char filename[PATH_MAX + 1], addr[18], str[38];
 	int i;
@@ -461,7 +461,7 @@ int write_link_key(bdaddr_t *local, bdaddr_t *peer, unsigned char *key, int type
 	return textfile_put(filename, addr, str);
 }
 
-int read_link_key(bdaddr_t *local, bdaddr_t *peer, unsigned char *key)
+int read_link_key(bdaddr_t *local, bdaddr_t *peer, unsigned char *key, uint8_t *type)
 {
 	char filename[PATH_MAX + 1], addr[18], tmp[3], *str;
 	int i;
@@ -479,6 +479,11 @@ int read_link_key(bdaddr_t *local, bdaddr_t *peer, unsigned char *key)
 		key[i] = (uint8_t) strtol(tmp, NULL, 16);
 	}
 
+	if (type) {
+		memcpy(tmp, str + 33, 2);
+		*type = (uint8_t) strtol(tmp, NULL, 10);
+	}
+
 	free(str);
 
 	return 0;
@@ -623,7 +628,7 @@ int write_trust(bdaddr_t *local, const char *addr, const char *service,
 	return ret;
 }
 
-gboolean read_trust(bdaddr_t *local, const char *addr, const char *service)
+gboolean read_trust(const bdaddr_t *local, const char *addr, const char *service)
 {
 	char filename[PATH_MAX + 1], *str;
 	GSList *services;
@@ -676,3 +681,27 @@ GSList *list_trusts(bdaddr_t *local, const char *service)
 
 	return list.trusts;
 }
+
+int write_device_profiles(bdaddr_t *src, bdaddr_t *dst, const char *profiles)
+{
+	char filename[PATH_MAX + 1], addr[18];
+
+	if (!profiles)
+		return -EINVAL;
+
+	create_filename(filename, PATH_MAX, src, "profiles");
+
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+
+	ba2str(dst, addr);
+	return textfile_put(filename, addr, profiles);
+}
+
+int delete_entry(bdaddr_t *src, const char *storage, const char *key)
+{
+	char filename[PATH_MAX + 1];
+
+	create_filename(filename, PATH_MAX, src, storage);
+
+	return textfile_del(filename, key);
+}
