@@ -29,6 +29,7 @@
 #include <errno.h>
 #include <unistd.h>
 #include <sys/stat.h>
+#include <sys/param.h>
 #include <netinet/in.h>
 
 #include <bluetooth/bluetooth.h>
@@ -36,20 +37,21 @@
 #include <bluetooth/hci_lib.h>
 #include <bluetooth/l2cap.h>
 #include <bluetooth/bnep.h>
+#include <bluetooth/sdp.h>
 
 #include <glib.h>
+#include <gdbus.h>
+
+#include "../hcid/dbus-common.h"
 
 #include "logging.h"
-#include "dbus.h"
-#include "dbus-helper.h"
 #include "textfile.h"
+#include "glib-helper.h"
 
 #include "error.h"
 #include "common.h"
 #include "connection.h"
 
-#define NETWORK_CONNECTION_INTERFACE "org.bluez.network.Connection"
-
 typedef enum {
 	CONNECTED,
 	CONNECTING,
@@ -77,6 +79,45 @@ struct __service_16 {
 
 static DBusConnection *connection = NULL;
 static const char *prefix = NULL;
+static GSList *connections = NULL;
+
+gint find_connection(gconstpointer a, gconstpointer b)
+{
+	const struct network_conn *nc = a;
+	const char *path = b;
+
+	return strcmp(nc->path, path);
+}
+
+static inline DBusMessage *not_supported(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+							"Not suported");
+}
+
+static inline DBusMessage *already_connected(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+						"Device already connected");
+}
+
+static inline DBusMessage *not_connected(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+						"Device not connected");
+}
+
+static inline DBusMessage *no_pending_connect(DBusMessage *msg)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+					"Device has no pending connect");
+}
+
+static inline DBusMessage *connection_attempt_failed(DBusMessage *msg, int err)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".ConnectionAttemptFailed",
+				err ? strerror(err) : "Connection attempt failed");
+}
 
 static gboolean bnep_watchdog_cb(GIOChannel *chan, GIOCondition cond,
 				gpointer data)
@@ -84,21 +125,25 @@ static gboolean bnep_watchdog_cb(GIOChannel *chan, GIOCondition cond,
 	struct network_conn *nc = data;
 
 	if (connection != NULL) {
-		dbus_connection_emit_signal(connection, nc->path,
+		g_dbus_emit_signal(connection, nc->path,
 						NETWORK_CONNECTION_INTERFACE,
 						"Disconnected",
 						DBUS_TYPE_INVALID);
 	}
+
 	info("%s disconnected", nc->dev);
+
+	bnep_if_down(nc->dev);
 	nc->state = DISCONNECTED;
 	memset(nc->dev, 0, 16);
 	strncpy(nc->dev, prefix, strlen(prefix));
 	g_io_channel_close(chan);
+
 	return FALSE;
 }
 
 static gboolean bnep_connect_cb(GIOChannel *chan, GIOCondition cond,
-				gpointer data)
+							gpointer data)
 {
 	struct network_conn *nc = data;
 	struct bnep_control_rsp *rsp;
@@ -159,17 +204,16 @@ static gboolean bnep_connect_cb(GIOChannel *chan, GIOCondition cond,
 	}
 
 	bnep_if_up(nc->dev, nc->id);
-	dbus_connection_emit_signal(connection, nc->path,
+	g_dbus_emit_signal(connection, nc->path,
 					NETWORK_CONNECTION_INTERFACE,
 					"Connected",
 					DBUS_TYPE_INVALID);
 
-	reply = dbus_message_new_method_return(nc->msg);
-
 	pdev = nc->dev;
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &pdev,
-					DBUS_TYPE_INVALID);
-	send_message_and_unref(connection, reply);
+
+	reply = g_dbus_create_reply(nc->msg, DBUS_TYPE_STRING, &pdev,
+							DBUS_TYPE_INVALID);
+	g_dbus_send_message(connection, reply);
 
 	nc->state = CONNECTED;
 
@@ -178,12 +222,15 @@ static gboolean bnep_connect_cb(GIOChannel *chan, GIOCondition cond,
 	g_io_add_watch(chan, G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 			(GIOFunc) bnep_watchdog_cb, nc);
 	return FALSE;
+
 failed:
 	if (nc->state != DISCONNECTED) {
 		nc->state = DISCONNECTED;
-		error_connection_attempt_failed(connection, nc->msg, EIO);
+		reply = connection_attempt_failed(nc->msg, EIO);
+		g_dbus_send_message(connection, reply);
 		g_io_channel_close(chan);
 	}
+
 	return FALSE;
 }
 
@@ -220,331 +267,176 @@ out:
 	return err;
 }
 
-static gboolean l2cap_connect_cb(GIOChannel *chan,
-			GIOCondition cond, gpointer data)
-{
-	struct network_conn *nc = data;
-	socklen_t len;
-	int sk, ret;
-
-	if (cond & G_IO_NVAL)
-		return FALSE;
-
-	if (cond & (G_IO_ERR | G_IO_HUP))
-		goto failed;
-
-	sk = g_io_channel_unix_get_fd(chan);
-
-	len = sizeof(ret);
-	if (getsockopt(sk, SOL_SOCKET, SO_ERROR, &ret, &len) < 0) {
-		error("getsockopt(SO_ERROR): %s (%d)", strerror(errno), errno);
-		goto failed;
-	}
-
-	if (ret != 0) {
-		error("connect(): %s (%d)", strerror(errno), errno);
-		goto failed;
-	}
-
-	if (bnep_connect(nc)) {
-		error("connect(): %s (%d)", strerror(errno), errno);
-		goto failed;
-	}
-
-	return FALSE;
-failed:
-	nc->state = DISCONNECTED;
-	error_connection_attempt_failed(connection, nc->msg, errno);
-	g_io_channel_close(chan);
-	return FALSE;
-}
-
-static int l2cap_connect(struct network_conn *nc)
-{
-	struct l2cap_options l2o;
-	struct sockaddr_l2 l2a;
-	socklen_t olen;
-	char addr[18];
-	GIOChannel *io;
-
-	ba2str(&nc->dst, addr);
-	info("Connecting to %s", addr);
-
-	/* Setup L2CAP options according to BNEP spec */
-	memset(&l2o, 0, sizeof(l2o));
-	olen = sizeof(l2o);
-	getsockopt(nc->sk, SOL_L2CAP, L2CAP_OPTIONS, &l2o, &olen);
-	l2o.imtu = l2o.omtu = BNEP_MTU;
-	setsockopt(nc->sk, SOL_L2CAP, L2CAP_OPTIONS, &l2o, sizeof(l2o));
-
-	memset(&l2a, 0, sizeof(l2a));
-	l2a.l2_family = AF_BLUETOOTH;
-	bacpy(&l2a.l2_bdaddr, &nc->src);
-
-	if (bind(nc->sk, (struct sockaddr *) &l2a, sizeof(l2a)) < 0) {
-		error("Bind failed. %s(%d)", strerror(errno), errno);
-		return -errno;
-	}
-
-	memset(&l2a, 0, sizeof(l2a));
-	l2a.l2_family = AF_BLUETOOTH;
-	bacpy(&l2a.l2_bdaddr, &nc->dst);
-	l2a.l2_psm = htobs(BNEP_PSM);
-
-	if (set_nonblocking(nc->sk) < 0) {
-		error("Set non blocking: %s (%d)", strerror(errno), errno);
-		return -errno;
-	}
-
-	io = g_io_channel_unix_new(nc->sk);
-	g_io_channel_set_close_on_unref(io, FALSE);
-
-	if (connect(nc->sk, (struct sockaddr *) &l2a, sizeof(l2a))) {
-		if (!(errno == EAGAIN || errno == EINPROGRESS)) {
-			error("Connect failed. %s(%d)", strerror(errno),
-					errno);
-			g_io_channel_close(io);
-			g_io_channel_unref(io);
-			return -errno;
-		}
-		g_io_add_watch(io, G_IO_OUT | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-				(GIOFunc) l2cap_connect_cb, nc);
-
-	} else {
-		l2cap_connect_cb(io, G_IO_OUT, nc);
-	}
-
-	g_io_channel_unref(io);
-	return 0;
-}
-
-static DBusHandlerResult get_adapter(DBusConnection *conn, DBusMessage *msg,
-					void *data)
+static void connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
+			const bdaddr_t *dst, gpointer data)
 {
 	struct network_conn *nc = data;
 	DBusMessage *reply;
+
+	if (err < 0) {
+		error("l2cap connect(): %s (%d)", strerror(-err), -err);
+		goto failed;
+	}
+
+	nc->sk = g_io_channel_unix_get_fd(chan);
+
+	err = bnep_connect(nc);
+	if (err < 0) {
+		error("bnep connect(): %s (%d)", strerror(-err), -err);
+		g_io_channel_close(chan);
+		g_io_channel_unref(chan);
+		goto failed;
+	}
+
+	return;
+
+failed:
+	nc->state = DISCONNECTED;
+
+	reply = connection_attempt_failed(nc->msg, -err);
+	g_dbus_send_message(connection, reply);
+}
+
+static DBusMessage *get_adapter(DBusConnection *conn, DBusMessage *msg,
+					void *data)
+{
+	struct network_conn *nc = data;
 	char addr[18];
 	const char *paddr = addr;
 
 	ba2str(&nc->src, addr);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &paddr,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &paddr,
+							DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult get_address(DBusConnection *conn, DBusMessage *msg,
+static DBusMessage *get_address(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
 	struct network_conn *nc = data;
-	DBusMessage *reply;
 	char addr[18];
 	const char *paddr = addr;
 
 	ba2str(&nc->dst, addr);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &paddr,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &paddr,
+							DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult get_uuid(DBusConnection *conn, DBusMessage *msg,
-					void *data)
-{
-	struct network_conn *nc = data;
-	const char *uuid;
-	DBusMessage *reply;
-
-	uuid = bnep_uuid(nc->id);
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &uuid,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
-}
-
-static DBusHandlerResult get_name(DBusConnection *conn, DBusMessage *msg,
-					void *data)
-{
-	struct network_conn *nc = data;
-	DBusMessage *reply;
-
-	if (!nc->name) {
-		error_failed(conn, msg, "Cannot find service name");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &nc->name,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
-
-}
-
-static DBusHandlerResult get_description(DBusConnection *conn,
+static DBusMessage *get_uuid(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct network_conn *nc = data;
-	DBusMessage *reply;
+	const char *uuid;
 
-	if (!nc->desc) {
-		error_failed(conn, msg, "Cannot find service description");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	uuid = bnep_uuid(nc->id);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &nc->desc,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &uuid,
+							DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult get_interface(DBusConnection *conn, DBusMessage *msg,
-					void *data)
+static DBusMessage *get_name(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	struct network_conn *nc = data;
+
+	if (!nc->name)
+		return not_supported(msg);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &nc->name,
+							DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *get_description(DBusConnection *conn,
+					DBusMessage *msg, void *data)
+{
+	struct network_conn *nc = data;
+
+	if (!nc->desc)
+		return not_supported(msg);
+
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &nc->desc,
+							DBUS_TYPE_INVALID);
+}
+
+static DBusMessage *get_interface(DBusConnection *conn,
+					DBusMessage *msg, void *data)
 {
 	struct network_conn *nc = data;
 	const char *pdev = nc->dev;
-	DBusMessage *reply;
 
-	if (nc->state != CONNECTED) {
-		error_failed(conn, msg, "Device not connected");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (nc->state != CONNECTED)
+		return not_connected(msg);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_message_append_args(reply, DBUS_TYPE_STRING, &pdev,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &pdev,
+						DBUS_TYPE_INVALID);
 }
 
 /* Connect and initiate BNEP session */
-static DBusHandlerResult connection_connect(DBusConnection *conn,
+static DBusMessage *connection_connect(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	struct network_conn *nc = data;
-	DBusError derr;
+	int err;
 
-	if (nc->state != DISCONNECTED) {
-		error_failed(conn, msg, "Device already connected");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (nc->state != DISCONNECTED)
+		return already_connected(msg);
 
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
-				DBUS_TYPE_INVALID)) {
-		error_invalid_arguments(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
-
-	nc->sk = socket(AF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP);
 	nc->state = CONNECTING;
-	if (nc->sk < 0) {
-		error("Cannot create L2CAP socket. %s(%d)", strerror(errno),
-				errno);
-		goto fail;
-	}
-
 	nc->msg = dbus_message_ref(msg);
-	if (l2cap_connect(nc) < 0) {
-		error("Connect failed. %s(%d)", strerror(errno), errno);
-		goto fail;
-	}
 
-	return DBUS_HANDLER_RESULT_HANDLED;
-fail:
-	if (nc->msg) {
+	err = bt_l2cap_connect(&nc->src, &nc->dst, BNEP_PSM, BNEP_MTU,
+							connect_cb, nc);
+	if (err < 0) {
+		error("Connect failed. %s(%d)", strerror(errno), errno);
 		dbus_message_unref(nc->msg);
 		nc->msg = NULL;
+		nc->state = DISCONNECTED;
+		return connection_attempt_failed(msg, -err);
 	}
-	nc->state = DISCONNECTED;
-	error_connection_attempt_failed(conn, msg, errno);
-	return DBUS_HANDLER_RESULT_HANDLED;
+
+	return NULL;
 }
 
-static DBusHandlerResult connection_cancel(DBusConnection *conn,
+static DBusMessage *connection_cancel(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	struct network_conn *nc = data;
-	DBusMessage *reply;
 
-	if (nc->state != CONNECTING) {
-		error_failed(conn, msg, "Device has no pending connect");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (nc->state != CONNECTING)
+		return no_pending_connect(msg);
 
 	close(nc->sk);
 	nc->state = DISCONNECTED;
 
-	reply = dbus_message_new_method_return(msg);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult connection_disconnect(DBusConnection *conn,
+static DBusMessage *connection_disconnect(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct network_conn *nc = data;
-	DBusMessage *reply;
 
-	if (nc->state != CONNECTED) {
-		error_failed(conn, msg, "Device not connected");
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+	if (nc->state != CONNECTED)
+		return not_connected(msg);
 
 	bnep_if_down(nc->dev);
 	bnep_kill_connection(&nc->dst);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult is_connected(DBusConnection *conn, DBusMessage *msg,
-					void *data)
+static DBusMessage *is_connected(DBusConnection *conn,
+				DBusMessage *msg, void *data)
 {
 	struct network_conn *nc = data;
-	DBusMessage *reply;
-	gboolean up;
+	gboolean up = (nc->state == CONNECTED);
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	up = (nc->state == CONNECTED);
-	dbus_message_append_args(reply, DBUS_TYPE_BOOLEAN, &up,
-					DBUS_TYPE_INVALID);
-
-	return send_message_and_unref(conn, reply);
+	return g_dbus_create_reply(msg, DBUS_TYPE_BOOLEAN, &up,
+						DBUS_TYPE_INVALID);
 }
 
-static DBusHandlerResult get_info(DBusConnection *conn,
+static DBusMessage *get_info(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct network_conn *nc = data;
@@ -556,8 +448,8 @@ static DBusHandlerResult get_info(DBusConnection *conn,
 	const char *paddr = raddr;
 
 	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+	if (reply == NULL)
+		return NULL;
 
 	dbus_message_iter_init_append(reply, &iter);
 
@@ -579,7 +471,7 @@ static DBusHandlerResult get_info(DBusConnection *conn,
 
 	dbus_message_iter_close_container(&iter, &dict);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
 static void connection_free(struct network_conn *nc)
@@ -600,39 +492,41 @@ static void connection_free(struct network_conn *nc)
 
 	if (nc->desc)
 		g_free(nc->desc);
-	
+
 	g_free(nc);
 	nc = NULL;
 }
 
-static void connection_unregister(DBusConnection *conn, void *data)
+static void connection_unregister(void *data)
 {
 	struct network_conn *nc = data;
 
 	info("Unregistered connection path:%s", nc->path);
 
+	connections = g_slist_remove(connections, nc);
 	connection_free(nc);
 }
 
-static DBusMethodVTable connection_methods[] = {
-	{ "GetAdapter",		get_adapter,		"",	"s"	},
-	{ "GetAddress",		get_address,		"",	"s"	},
-	{ "GetUUID",		get_uuid,		"",	"s"	},
-	{ "GetName",		get_name,		"",	"s"	},
-	{ "GetDescription",	get_description,	"",	"s"	},
-	{ "GetInterface",	get_interface,		"",	"s"	},
-	{ "Connect",		connection_connect,	"",	"s"	},
-	{ "CancelConnect",	connection_cancel,	"",	""	},
-	{ "Disconnect",		connection_disconnect,	"",	""	},
-	{ "IsConnected",	is_connected,		"",	"b"	},
-	{ "GetInfo",		get_info,		"",	"a{sv}" },
-	{ NULL, NULL, NULL, NULL }
+static GDBusMethodTable connection_methods[] = {
+	{ "GetAdapter",		"",	"s",	get_adapter		},
+	{ "GetAddress",		"",	"s",	get_address		},
+	{ "GetUUID",		"",	"s",	get_uuid		},
+	{ "GetName",		"",	"s",	get_name		},
+	{ "GetDescription",	"",	"s",	get_description		},
+	{ "GetInterface",	"",	"s",	get_interface		},
+	{ "Connect",		"",	"s",	connection_connect,
+						G_DBUS_METHOD_FLAG_ASYNC },
+	{ "CancelConnect",	"",	"",	connection_cancel	},
+	{ "Disconnect",		"",	"",	connection_disconnect	},
+	{ "IsConnected",	"",	"b",	is_connected		},
+	{ "GetInfo",		"",	"a{sv}",get_info		},
+	{ }
 };
 
-static DBusSignalVTable connection_signals[] = {
+static GDBusSignalTable connection_signals[] = {
 	{ "Connected",		""	},
 	{ "Disconnected",	""	},
-	{ NULL, NULL }
+	{ }
 };
 
 int connection_register(const char *path, bdaddr_t *src, bdaddr_t *dst,
@@ -652,20 +546,13 @@ int connection_register(const char *path, bdaddr_t *src, bdaddr_t *dst,
 
 	nc = g_new0(struct network_conn, 1);
 
-	/* register path */
-	if (!dbus_connection_create_object_path(connection, path, nc,
-						connection_unregister)) {
-		connection_free(nc);
-		return -1;
-	}
-
-	if (!dbus_connection_register_interface(connection, path,
-						NETWORK_CONNECTION_INTERFACE,
-						connection_methods,
-						connection_signals, NULL)) {
+	if (g_dbus_register_interface(connection, path,
+					NETWORK_CONNECTION_INTERFACE,
+					connection_methods,
+					connection_signals, NULL,
+					nc, connection_unregister) == FALSE) {
 		error("D-Bus failed to register %s interface",
 				NETWORK_CONNECTION_INTERFACE);
-		dbus_connection_destroy_object_path(connection, path);
 		return -1;
 	}
 
@@ -680,6 +567,8 @@ int connection_register(const char *path, bdaddr_t *src, bdaddr_t *dst,
 	strncpy(nc->dev, prefix, strlen(prefix));
 	nc->state = DISCONNECTED;
 
+	connections = g_slist_append(connections, nc);
+
 	info("Registered connection path:%s", path);
 
 	return 0;
@@ -693,11 +582,13 @@ int connection_store(const char *path, gboolean default_path)
 	char filename[PATH_MAX + 1];
 	char src_addr[18], dst_addr[18];
 	int len, err;
+	GSList *l;
 
-	if (!dbus_connection_get_object_user_data(connection,
-				path, (void *) &nc))
+	l = g_slist_find_custom(connections, path, find_connection);
+	if (!l)
 		return -ENOENT;
 
+	nc = l->data;
 	if (!nc->name || !nc->desc)
 		return -EINVAL;
 
@@ -729,11 +620,13 @@ int connection_find_data(const char *path, const char *pattern)
 	struct network_conn *nc;
 	char addr[18], key[32];
 	const char *role;
+	GSList *l;
 
-	if (!dbus_connection_get_object_user_data(connection,
-		path, (void *) &nc))
+	l = g_slist_find_custom(connections, path, find_connection);
+	if (!l)
 		return -1;
 
+	nc = l->data;
 	if (strcasecmp(pattern, nc->dev) == 0)
 		return 0;
 
@@ -757,11 +650,14 @@ int connection_find_data(const char *path, const char *pattern)
 gboolean connection_has_pending(const char *path)
 {
 	struct network_conn *nc;
+	GSList *l;
 
-	if (!dbus_connection_get_object_user_data(connection,
-				path, (void *) &nc))
+	l = g_slist_find_custom(connections, path, find_connection);
+	if (!l)
 		return FALSE;
 
+	nc = l->data;
+
 	return (nc->state == CONNECTING);
 }
 
@@ -773,11 +669,14 @@ int connection_remove_stored(const char *path)
 	char filename[PATH_MAX + 1];
 	char src_addr[18], dst_addr[18];
 	int err;
+	GSList *l;
 
-	if (!dbus_connection_get_object_user_data(connection,
-				path, (void *) &nc))
+	l = g_slist_find_custom(connections, path, find_connection);
+	if (!l)
 		return -ENOENT;
 
+	nc = l->data;
+
 	ba2str(&nc->dst, dst_addr);
 	role = bnep_name(nc->id);
 	snprintf(key, 32, "%s#%s", dst_addr, role);
@@ -793,11 +692,14 @@ int connection_remove_stored(const char *path)
 gboolean connection_is_connected(const char *path)
 {
 	struct network_conn *nc;
+	GSList *l;
 
-	if (!dbus_connection_get_object_user_data(connection,
-				path, (void *) &nc))
+	l = g_slist_find_custom(connections, path, find_connection);
+	if (!l)
 		return FALSE;
 
+	nc = l->data;
+
 	return (nc->state == CONNECTED);
 }
 
