@@ -29,9 +29,8 @@
 #include <unistd.h>
 #include <stdlib.h>
 #include <errno.h>
-#include <sys/socket.h>
-#include <sys/ioctl.h>
 #include <sys/stat.h>
+#include <sys/param.h>
 #include <net/if.h>
 
 #include <bluetooth/bluetooth.h>
@@ -45,12 +44,16 @@
 #include <netinet/in.h>
 
 #include <glib.h>
+#include <gdbus.h>
+
+#include "../hcid/dbus-common.h"
 
 #include "logging.h"
-#include "dbus.h"
 #include "error.h"
 #include "textfile.h"
-#include "dbus-helper.h"
+#include "dbus-service.h"
+#include "sdpd.h"
+#include "glib-helper.h"
 
 #define NETWORK_SERVER_INTERFACE "org.bluez.network.Server"
 #define SETUP_TIMEOUT		1000
@@ -66,9 +69,12 @@ struct setup_session {
 	uint16_t	dst_role;	/* Destination role */
 	uint16_t	src_role;	/* Source role */
 	int		nsk;		/* L2CAP socket */
-	int		attempts;	/* Setup msg received */
-	guint		watch;		/* BNEP setup watch */
-	guint		timeout;	/* Max setup time */
+	guint		watch;		/* BNEP socket watch */
+};
+
+struct timeout {
+	guint	id;		/* Timeout id */
+	guint	watch;		/* BNEP socket watch */
 };
 
 /* Main server structure */
@@ -86,10 +92,57 @@ struct network_server {
 
 static GIOChannel *bnep_io = NULL;
 static DBusConnection *connection = NULL;
-static GSList *setup_sessions = NULL;
+static struct setup_session *setup = NULL;
+static GSList *servers = NULL;
 static const char *prefix = NULL;
 static gboolean security = TRUE;
 
+gint find_server(gconstpointer a, gconstpointer b)
+{
+	const struct network_server *ns = a;
+	const char *path = b;
+
+	return strcmp(ns->path, path);
+}
+
+static struct setup_session *setup_session_new(gchar *address,
+		uint16_t dst_role, uint16_t src_role, int nsk, guint watch)
+{
+	struct setup_session *setup;
+
+	setup = g_new0(struct setup_session, 1);
+	setup->address = g_strdup(address);
+	setup->dst_role = dst_role;
+	setup->src_role = src_role;
+	setup->nsk = nsk;
+	setup->watch = watch;
+
+	return setup;
+}
+
+static void setup_session_free(struct setup_session *setup)
+{
+	g_source_remove(setup->watch);
+	g_free(setup->address);
+	g_free(setup);
+}
+
+static struct network_server *server_find(bdaddr_t *src, uint16_t role)
+{
+	struct network_server *ns;
+	GSList *l;
+
+	for (l = servers; l; l = l->next) {
+		ns = l->data;
+		if (bacmp(&ns->src, src) != 0)
+			continue;
+		if (ns->id == role)
+			return ns;
+	}
+
+	return NULL;
+}
+
 static int store_property(bdaddr_t *src, uint16_t id,
 			const char *key, const char *value)
 {
@@ -107,17 +160,6 @@ static int store_property(bdaddr_t *src, uint16_t id,
 	return textfile_put(filename, key, value);
 }
 
-static void setup_free(struct setup_session *s)
-{
-	g_free(s->address);
-	g_free(s);
-}
-
-static int setup_cmp(const struct setup_session *s, const char *addr)
-{
-	return strcmp(s->address, addr);
-}
-
 static void add_lang_attr(sdp_record_t *r)
 {
 	sdp_lang_attr_t base_lang;
@@ -132,8 +174,7 @@ static void add_lang_attr(sdp_record_t *r)
 	sdp_list_free(langs, 0);
 }
 
-static int create_server_record(sdp_buf_t *buf, const char *name,
-					uint16_t id)
+static sdp_record_t *server_record_new(const char *name, uint16_t id)
 {
 	sdp_list_t *svclass, *pfseq, *apseq, *root, *aproto;
 	uuid_t root_uuid, pan, l2cap, bnep;
@@ -145,60 +186,64 @@ static int create_server_record(sdp_buf_t *buf, const char *name,
 	uint16_t net_access_type = 0xfffe;
 	uint32_t max_net_access_rate = 0;
 	const char *desc = "BlueZ PAN service";
-	sdp_record_t record;
-	int ret;
+	sdp_record_t *record;
 
-	memset(&record, 0, sizeof(sdp_record_t));
+	record = sdp_record_alloc();
+	if (!record)
+		return NULL;
+
+	record->attrlist = NULL;
+	record->pattern = NULL;
 
 	switch (id) {
 	case BNEP_SVC_NAP:
 		sdp_uuid16_create(&pan, NAP_SVCLASS_ID);
 		svclass = sdp_list_append(NULL, &pan);
-		sdp_set_service_classes(&record, svclass);
+		sdp_set_service_classes(record, svclass);
 
 		sdp_uuid16_create(&profile[0].uuid, NAP_PROFILE_ID);
 		profile[0].version = 0x0100;
 		pfseq = sdp_list_append(NULL, &profile[0]);
-		sdp_set_profile_descs(&record, pfseq);
+		sdp_set_profile_descs(record, pfseq);
 
-		sdp_set_info_attr(&record, name, NULL, desc);
+		sdp_set_info_attr(record, name, NULL, desc);
 
-		sdp_attr_add_new(&record, SDP_ATTR_NET_ACCESS_TYPE,
+		sdp_attr_add_new(record, SDP_ATTR_NET_ACCESS_TYPE,
 					SDP_UINT16, &net_access_type);
-		sdp_attr_add_new(&record, SDP_ATTR_MAX_NET_ACCESSRATE,
+		sdp_attr_add_new(record, SDP_ATTR_MAX_NET_ACCESSRATE,
 					SDP_UINT32, &max_net_access_rate);
 		break;
 	case BNEP_SVC_GN:
 		sdp_uuid16_create(&pan, GN_SVCLASS_ID);
 		svclass = sdp_list_append(NULL, &pan);
-		sdp_set_service_classes(&record, svclass);
+		sdp_set_service_classes(record, svclass);
 
 		sdp_uuid16_create(&profile[0].uuid, GN_PROFILE_ID);
 		profile[0].version = 0x0100;
 		pfseq = sdp_list_append(NULL, &profile[0]);
-		sdp_set_profile_descs(&record, pfseq);
+		sdp_set_profile_descs(record, pfseq);
 
-		sdp_set_info_attr(&record, name, NULL, desc);
+		sdp_set_info_attr(record, name, NULL, desc);
 		break;
 	case BNEP_SVC_PANU:
 		sdp_uuid16_create(&pan, PANU_SVCLASS_ID);
 		svclass = sdp_list_append(NULL, &pan);
-		sdp_set_service_classes(&record, svclass);
+		sdp_set_service_classes(record, svclass);
 
 		sdp_uuid16_create(&profile[0].uuid, PANU_PROFILE_ID);
 		profile[0].version = 0x0100;
 		pfseq = sdp_list_append(NULL, &profile[0]);
-		sdp_set_profile_descs(&record, pfseq);
+		sdp_set_profile_descs(record, pfseq);
 
-		sdp_set_info_attr(&record, name, NULL, desc);
+		sdp_set_info_attr(record, name, NULL, desc);
 		break;
 	default:
-		return -1;
+		return NULL;
 	}
 
 	sdp_uuid16_create(&root_uuid, PUBLIC_BROWSE_GROUP);
 	root = sdp_list_append(NULL, &root_uuid);
-	sdp_set_browse_groups(&record, root);
+	sdp_set_browse_groups(record, root);
 
 	sdp_uuid16_create(&l2cap, L2CAP_UUID);
 	proto[0] = sdp_list_append(NULL, &l2cap);
@@ -234,18 +279,13 @@ static int create_server_record(sdp_buf_t *buf, const char *name,
 	apseq = sdp_list_append(apseq, proto[1]);
 
 	aproto = sdp_list_append(NULL, apseq);
-	sdp_set_access_protos(&record, aproto);
+	sdp_set_access_protos(record, aproto);
 
-	add_lang_attr(&record);
+	add_lang_attr(record);
 
-	sdp_attr_add_new(&record, SDP_ATTR_SECURITY_DESC,
+	sdp_attr_add_new(record, SDP_ATTR_SECURITY_DESC,
 				SDP_UINT16, &security_desc);
 
-	if (sdp_gen_record_pdu(&record, buf) < 0)
-		ret = -1;
-	else
-		ret = 0;
-
 	sdp_data_free(p);
 	sdp_data_free(v);
 	sdp_list_free(apseq, NULL);
@@ -255,85 +295,40 @@ static int create_server_record(sdp_buf_t *buf, const char *name,
 	sdp_list_free(proto[1], NULL);
 	sdp_list_free(svclass, NULL);
 	sdp_list_free(pfseq, NULL);
-	sdp_list_free(record.attrlist, (sdp_free_func_t) sdp_data_free);
-	sdp_list_free(record.pattern, free);
 
-	return ret;
+	return record;
 }
 
-static ssize_t send_bnep_ctrl_rsp(int sk, uint16_t response)
+static ssize_t send_bnep_ctrl_rsp(int sk, uint16_t val)
 {
 	struct bnep_control_rsp rsp;
 
 	rsp.type = BNEP_CONTROL;
 	rsp.ctrl = BNEP_SETUP_CONN_RSP;
-	rsp.resp = htons(response);
+	rsp.resp = htons(val);
 
 	return send(sk, &rsp, sizeof(rsp), 0);
 }
 
-static void cancel_authorization(struct setup_session *s)
+static int server_connadd(struct network_server *ns, int nsk,
+			const gchar *address, uint16_t dst_role)
 {
-	DBusMessage *msg;
-	const char *uuid;
-
-	msg = dbus_message_new_method_call("org.bluez", "/org/bluez",
-						"org.bluez.Database",
-						"CancelAuthorizationRequest");
-	if (!msg) {
-		error("Unable to allocate new method call");
-		return;
-	}
-
-	uuid = bnep_uuid(s->dst_role);
-	dbus_message_append_args(msg,
-			DBUS_TYPE_STRING, &s->address,
-			DBUS_TYPE_STRING, &uuid,
-			DBUS_TYPE_INVALID);
-
-	send_message_and_unref(connection, msg);
-}
-
-static void authorization_callback(DBusPendingCall *pcall, void *data)
-{
-	struct setup_session *s = data;
-	struct network_server *ns = NULL;
-	DBusMessage *reply = dbus_pending_call_steal_reply(pcall);
-	char path[MAX_PATH_LENGTH], devname[16];
-	uint16_t response = BNEP_CONN_NOT_ALLOWED;
-	DBusError derr;
+	char devname[16];
 	const char *bridge;
-
-	if (!g_slist_find(setup_sessions, s)) {
-		dbus_message_unref(reply);
-		return;
-	}
-
-	snprintf(path, MAX_PATH_LENGTH, NETWORK_PATH"/%s", bnep_name(s->dst_role));
-	dbus_connection_get_object_user_data(connection, path, (void *) &ns);
+	int err;
 
 	/* Server can be disabled in the meantime */
-	if (ns == NULL || ns->enable == FALSE)
-		goto failed;
-
-	dbus_error_init(&derr);
-	if (dbus_set_error_from_message(&derr, reply)) {
-		error("Access denied: %s", derr.message);
-		if (dbus_error_has_name(&derr, DBUS_ERROR_NO_REPLY)) {
-			debug("Canceling authorization request");
-			cancel_authorization(s);
-		}
-		dbus_error_free(&derr);
-		goto failed;
-	}
+	if (ns->enable == FALSE)
+		return -EPERM;
 
 	memset(devname, 0, 16);
 	strncpy(devname, prefix, strlen(prefix));
 
-	if (bnep_connadd(s->nsk, s->dst_role, devname) < 0)
-		goto failed;
+	err = bnep_connadd(nsk, dst_role, devname);
+	if (err < 0)
+		return err;
 
-	info("Authorization succedded. New connection: %s", devname);
+	info("Added new connection: %s", devname);
 
 	bridge = bridge_get_name(ns->id);
 	if (bridge) {
@@ -341,84 +336,66 @@ static void authorization_callback(DBusPendingCall *pcall, void *data)
 			error("Can't add %s to the bridge %s: %s(%d)",
 					devname, bridge, strerror(errno),
 					errno);
-			goto failed;
+			return -EPERM;
 		}
 
 		bnep_if_up(devname, 0);
 	} else
 		bnep_if_up(devname, ns->id);
 
-	response = BNEP_SUCCESS;
-
-	ns->clients = g_slist_append(ns->clients, g_strdup(s->address));
-
-failed:
-	send_bnep_ctrl_rsp(s->nsk, response);
-	dbus_message_unref(reply);
-}
-
-static void setup_watch_destroy(void *data)
-{
-	struct setup_session *s;
-	GSList *l;
-
-	/*
-	 * Remote initiated: socket HUP
-	 * Authorization: denied/accepted
-	 */
-	l = g_slist_find(setup_sessions, data);
-	if (!l)
-		return;
-
-	s = l->data;
-
-	setup_sessions = g_slist_remove(setup_sessions, s);
-
-	/* Remove active watches */
-	if (s->watch)
-		g_source_remove(s->watch);
-	if (s->timeout)
-		g_source_remove(s->timeout);
-	setup_free(s);
-}
-
-static int authorize_connection(struct setup_session *s)
-{
-	DBusMessage *msg;
-	DBusPendingCall *pending;
-	const char *uuid;
-
-	msg = dbus_message_new_method_call("org.bluez", "/org/bluez",
-				"org.bluez.Database", "RequestAuthorization");
-	if (!msg) {
-		error("Unable to allocat new RequestAuthorization method call");
-		return -ENOMEM;
-	}
-
-	uuid = bnep_uuid(s->dst_role);
-	debug("Requesting authorization for %s UUID:%s", s->address, uuid);
-
-	dbus_message_append_args(msg,
-			DBUS_TYPE_STRING, &s->address,
-			DBUS_TYPE_STRING, &uuid,
-			DBUS_TYPE_INVALID);
-
-	if (dbus_connection_send_with_reply(connection,
-				msg, &pending, -1) == FALSE) {
-		error("Sending of authorization request failed");
-		dbus_message_unref(msg);
-		return -EACCES;
-	}
-
-	dbus_pending_call_set_notify(pending,
-			authorization_callback, s, setup_watch_destroy);
-	dbus_pending_call_unref(pending);
-	dbus_message_unref(msg);
+	ns->clients = g_slist_append(ns->clients, g_strdup(address));
 
 	return 0;
 }
 
-static uint16_t inline chk_role(uint16_t dst_role, uint16_t src_role)
+static void req_auth_cb(DBusError *derr, void *user_data)
+{
+	struct network_server *ns = user_data;
+	uint16_t val;
+
+	if (!setup) {
+		info("Authorization cancelled: Client exited");
+		return;
+	}
+
+	if (derr) {
+		error("Access denied: %s", derr->message);
+		if (dbus_error_has_name(derr, DBUS_ERROR_NO_REPLY)) {
+			bdaddr_t dst;
+			str2ba(setup->address, &dst);
+			service_cancel_auth(&ns->src, &dst);
+		}
+		val = BNEP_CONN_NOT_ALLOWED;
+		goto done;
+	}
+
+	if (server_connadd(ns, setup->nsk,
+			setup->address, setup->dst_role) < 0)
+		val = BNEP_CONN_NOT_ALLOWED;
+	else
+		val = BNEP_SUCCESS;
+
+done:
+	send_bnep_ctrl_rsp(setup->nsk, val);
+	setup_session_free(setup);
+	setup = NULL;
+}
+
+static int authorize_connection(struct network_server *ns, const char *address)
+{
+	const char *uuid;
+	bdaddr_t dst;
+	int ret_val;
+
+	uuid =  bnep_uuid(ns->id);
+	str2ba(address, &dst);
+
+	ret_val = service_req_auth(&ns->src, &dst, uuid, req_auth_cb, ns);
+
+	return ret_val;
+}
+
+static uint16_t inline bnep_setup_chk(uint16_t dst_role, uint16_t src_role)
 {
 	/* Allowed PAN Profile scenarios */
 	switch (dst_role) {
@@ -439,249 +416,192 @@ static uint16_t inline chk_role(uint16_t dst_role, uint16_t src_role)
 	return BNEP_CONN_INVALID_DST;
 }
 
-static gboolean connect_setup_event(GIOChannel *chan,
-					GIOCondition cond, gpointer data)
+static uint16_t bnep_setup_decode(struct bnep_setup_conn_req *req,
+				uint16_t *dst_role, uint16_t *src_role)
 {
-	struct setup_session *s = data;
-	struct network_server *ns = NULL;
-	struct bnep_setup_conn_req *req;
-	unsigned char pkt[BNEP_MTU];
-	char path[MAX_PATH_LENGTH];
-	uint16_t response;
-	uint8_t *pservice;
-	ssize_t r;
-	int sk;
+	uint8_t *dest, *source;
+
+	dest = req->service;
+	source = req->service + req->uuid_size;
+
+	switch (req->uuid_size) {
+	case 2: /* UUID16 */
+		*dst_role = ntohs(bt_get_unaligned((uint16_t *) dest));
+		*src_role = ntohs(bt_get_unaligned((uint16_t *) source));
+		break;
+	case 4: /* UUID32 */
+	case 16: /* UUID128 */
+		*dst_role = ntohl(bt_get_unaligned((uint32_t *) dest));
+		*src_role = ntohl(bt_get_unaligned((uint32_t *) source));
+		break;
+	default:
+		return BNEP_CONN_INVALID_SVC;
+	}
+
+	return 0;
+}
+
+static gboolean bnep_setup(GIOChannel *chan,
+			GIOCondition cond, gpointer user_data)
+{
+	struct timeout *to = user_data;
+	struct network_server *ns;
+	uint8_t packet[BNEP_MTU];
+	struct bnep_setup_conn_req *req = (void *) packet;
+	struct sockaddr_l2 sa;
+	socklen_t size;
+	char address[18];
+	uint16_t rsp, src_role, dst_role;
+	int n, sk;
 
 	if (cond & G_IO_NVAL)
 		return FALSE;
 
 	if (cond & (G_IO_ERR | G_IO_HUP)) {
 		error("Hangup or error on BNEP socket");
-		/* If there is a pending authorization */
-		if (s->attempts)
-			cancel_authorization(s);
 		return FALSE;
 	}
 
 	sk = g_io_channel_unix_get_fd(chan);
-	memset(pkt, 0, sizeof(pkt));
-	r = recv(sk, pkt, sizeof(pkt) - 1, 0);
 
-	req = (struct bnep_setup_conn_req *) pkt;
-	/*
-	 * FIXME: According to BNEP SPEC the UUID size can be
-	 * 2-16 bytes. Currently only 2 bytes size is supported
-	 */
-	if (req->uuid_size != 2 || r != (sizeof(*req) + req->uuid_size * 2)) {
-		error("Invalid BNEP packet size");
-		response = BNEP_CONN_INVALID_SVC;
-		goto reply;
-	}
-
-	if (req->type != BNEP_CONTROL || req->ctrl != BNEP_SETUP_CONN_REQ) {
-		error("Invalid BNEP control packet content");
+	/* Reading BNEP_SETUP_CONNECTION_REQUEST_MSG */
+	n = read(sk, packet, sizeof(packet));
+	if (n < 0) {
+		error("read(): %s(%d)", strerror(errno), errno);
 		return FALSE;
 	}
 
-	pservice = req->service;
-	/* Getting destination service: considering 2 bytes size */
-	s->dst_role = ntohs(bt_get_unaligned((uint16_t *) pservice));
-	pservice += req->uuid_size;
-	/* Getting source service: considering 2 bytes size */
-	s->src_role = ntohs(bt_get_unaligned((uint16_t *) pservice));
+	if (req->type != BNEP_CONTROL || req->ctrl != BNEP_SETUP_CONN_REQ)
+		return FALSE;
 
-	response = chk_role(s->src_role, s->dst_role);
-	if (response)
+	rsp = bnep_setup_decode(req, &dst_role, &src_role);
+	if (rsp)
 		goto reply;
 
-	snprintf(path, MAX_PATH_LENGTH, NETWORK_PATH"/%s", bnep_name(s->dst_role));
-	dbus_connection_get_object_user_data(connection, path, (void *) &ns);
+	rsp = bnep_setup_chk(dst_role, src_role);
+	if (rsp)
+		goto reply;
 
-	if (ns == NULL || ns->enable == FALSE) {
-		response = BNEP_CONN_NOT_ALLOWED;
+	size = sizeof(sa);
+	if (getsockname(sk, (struct sockaddr *) &sa, &size) < 0) {
+		rsp = BNEP_CONN_NOT_ALLOWED;
 		goto reply;
 	}
 
-	if (s->timeout) {
-		g_source_remove(s->timeout);
-		s->timeout = 0;
-	}
-
-	if (++s->attempts > MAX_SETUP_ATTEMPTS) {
-		/* Retransmission */
-		response = BNEP_CONN_NOT_ALLOWED;
+	ba2str(&sa.l2_bdaddr, address);
+	ns = server_find(&sa.l2_bdaddr, dst_role);
+	if (!ns || ns->enable == FALSE) {
+		error("Server unavailable: %s (0x%x)", address, dst_role);
+		rsp = BNEP_CONN_NOT_ALLOWED;
 		goto reply;
 	}
 
+	if (getpeername(sk, (struct sockaddr *) &sa, &size) < 0) {
+		rsp = BNEP_CONN_NOT_ALLOWED;
+		goto reply;
+	}
+
+	ba2str(&sa.l2_bdaddr, address);
+
+	if (setup) {
+		error("Connection rejected: Pending authorization");
+		rsp = BNEP_CONN_NOT_ALLOWED;
+		goto reply;
+	}
+
+	setup = setup_session_new(address, dst_role, src_role, sk, to->watch);
+
 	/* Wait authorization before reply success */
-	if (authorize_connection(s) < 0) {
-		response = BNEP_CONN_NOT_ALLOWED;
+	if (authorize_connection(ns, address) < 0) {
+		setup_session_free(setup);
+		setup = NULL;
+		rsp = BNEP_CONN_NOT_ALLOWED;
 		goto reply;
 	}
 
+	g_source_remove(to->id);
+	to->id = 0;
+
 	return TRUE;
+
 reply:
-	send_bnep_ctrl_rsp(sk, response);
+	send_bnep_ctrl_rsp(sk, rsp);
+
 	return FALSE;
 }
 
-static gboolean setup_timeout(void *data)
+static void setup_destroy(void *user_data)
 {
-	setup_watch_destroy(data);
+	struct timeout *to = user_data;
+
+	if (to->id)
+		g_source_remove(to->id);
+
+	g_free(to);
+}
+
+static gboolean timeout_cb(void *user_data)
+{
+	struct timeout *to = user_data;
+
+	to->id = 0;
+	g_source_remove(to->watch);
+
 	return FALSE;
 }
 
-static gboolean connect_event(GIOChannel *chan,
-				GIOCondition cond, gpointer data)
+static void connect_event(GIOChannel *chan, int err, const bdaddr_t *src,
+				const bdaddr_t *dst, gpointer user_data)
 {
-	struct sockaddr_l2 addr;
-	struct setup_session *s;
-	GIOChannel *io;
-	socklen_t addrlen;
-	char peer[18];
-	bdaddr_t dst;
-	unsigned short psm;
-	int sk, nsk;
+	struct timeout *to;
 
-	if (cond & G_IO_NVAL)
-		return FALSE;
-
-	if (cond & (G_IO_ERR | G_IO_HUP)) {
-		error("Hangup or error on L2CAP socket PSM 15");
-		g_io_channel_close(chan);
-		return FALSE;
+	if (err < 0) {
+		error("accept(): %s(%d)", strerror(errno), errno);
+		return;
 	}
 
-	sk = g_io_channel_unix_get_fd(chan);
+	g_io_channel_set_close_on_unref(chan, TRUE);
 
-	memset(&addr, 0, sizeof(addr));
-	addrlen = sizeof(addr);
+	/*
+	 * BNEP_SETUP_CONNECTION_REQUEST_MSG shall be received and
+	 * user shall authorize the incomming connection before
+	 * the time expires.
+	 */
+	to = g_malloc0(sizeof(struct timeout));
+	to->id = g_timeout_add(SETUP_TIMEOUT, timeout_cb, to);
+	to->watch = g_io_add_watch_full(chan, G_PRIORITY_DEFAULT,
+				G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+				bnep_setup, to, setup_destroy);
+	g_io_channel_unref(chan);
 
-	nsk = accept(sk, (struct sockaddr *) &addr, &addrlen);
-	if (nsk < 0)
-		return TRUE;
-
-	bacpy(&dst, &addr.l2_bdaddr);
-	psm = btohs(addr.l2_psm);
-	ba2str(&dst, peer);
-
-	info("Connection from: %s on PSM %d", peer, psm);
-
-	if (g_slist_find_custom(setup_sessions, peer,
-				(GCompareFunc) setup_cmp)) {
-		error("Pending connection setup session");
-		close(nsk);
-		return TRUE;
-	}
-
-	s = g_new0(struct setup_session, 1);
-	s->address = g_strdup(peer);
-	s->nsk = nsk;
-
-	io = g_io_channel_unix_new(nsk);
-	g_io_channel_set_close_on_unref(io, TRUE);
-	/* New watch for BNEP setup */
-	s->watch = g_io_add_watch_full(io, G_PRIORITY_DEFAULT,
-			G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-			connect_setup_event, s, &setup_watch_destroy);
-	g_io_channel_unref(io);
-
-	/* Remove the timeout at the first valid msg */
-	s->timeout = g_timeout_add(SETUP_TIMEOUT, setup_timeout, s);
-
-	setup_sessions = g_slist_append(setup_sessions, s);
-
-	return TRUE;
+	return;
 }
 
 int server_init(DBusConnection *conn, const char *iface_prefix,
 		gboolean secure)
 {
-	struct l2cap_options l2o;
-	struct sockaddr_l2 l2a;
-	socklen_t olen;
-	int sk, lm, err;
-
-	/* Create L2CAP socket and bind it to PSM BNEP */
-	sk = socket(AF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP);
-	if (sk < 0) {
-		err = errno;
-		error("Cannot create L2CAP socket. %s(%d)",
-					strerror(err), err);
-		return -err;
-	}
-
-	memset(&l2a, 0, sizeof(l2a));
-	l2a.l2_family = AF_BLUETOOTH;
-	bacpy(&l2a.l2_bdaddr, BDADDR_ANY);
-	l2a.l2_psm = htobs(BNEP_PSM);
-
-	if (bind(sk, (struct sockaddr *) &l2a, sizeof(l2a))) {
-		err = errno;
-		error("Bind failed. %s(%d)", strerror(err), err);
-		goto fail;
-	}
-
-	/* Setup L2CAP options according to BNEP spec */
-	memset(&l2o, 0, sizeof(l2o));
-	olen = sizeof(l2o);
-	if (getsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &l2o, &olen) < 0) {
-		err = errno;
-		error("Failed to get L2CAP options. %s(%d)",
-					strerror(err), err);
-		goto fail;
-	}
-
-	l2o.imtu = l2o.omtu = BNEP_MTU;
-	if (setsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &l2o, sizeof(l2o)) < 0) {
-		err = errno;
-		error("Failed to set L2CAP options. %s(%d)",
-					strerror(err), err);
-		goto fail;
-	}
+	int lm;
 
 	lm = secure ? L2CAP_LM_SECURE : 0;
-	if (lm && setsockopt(sk, SOL_L2CAP, L2CAP_LM, &lm, sizeof(lm)) < 0) {
-		err = errno;
-		error("Failed to set link mode. %s(%d)",
-					strerror(err), err);
-		goto fail;
-	}
+
 	security = secure;
-
-	if (listen(sk, 1) < 0) {
-		err = errno;
-		error("Listen failed. %s(%d)", strerror(err), err);
-		goto fail;
-	}
-
 	connection = dbus_connection_ref(conn);
 	prefix = iface_prefix;
 
-	bnep_io = g_io_channel_unix_new(sk);
+	bnep_io = bt_l2cap_listen(BDADDR_ANY, BNEP_PSM, BNEP_MTU, lm,
+			connect_event, NULL);
+	if (!bnep_io)
+		return -1;
 	g_io_channel_set_close_on_unref(bnep_io, FALSE);
-	g_io_add_watch(bnep_io, G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-							connect_event, NULL);
 
 	if (bridge_create(BNEP_SVC_GN) < 0)
 		error("Can't create GN bridge");
 
 	return 0;
-fail:
-
-	close(sk);
-	errno = err;
-	return -err;
 }
 
 void server_exit()
 {
-	if (setup_sessions) {
-		g_slist_foreach(setup_sessions, (GFunc) setup_free, NULL);
-		g_slist_free(setup_sessions);
-		setup_sessions = NULL;
-	}
-
 	if (bnep_io != NULL) {
 		g_io_channel_close(bnep_io);
 		g_io_channel_unref(bnep_io);
@@ -695,137 +615,29 @@ void server_exit()
 	connection = NULL;
 }
 
-static uint32_t add_server_record(struct network_server *ns)
+static uint32_t register_server_record(struct network_server *ns)
 {
-	DBusMessage *msg, *reply;
-	DBusError derr;
-	dbus_uint32_t rec_id;
-	sdp_buf_t buf;
+	sdp_record_t *record;
 
-	msg = dbus_message_new_method_call("org.bluez", "/org/bluez",
-				"org.bluez.Database", "AddServiceRecord");
-	if (!msg) {
-		error("Can't allocate new method call");
-		return 0;
-	}
-
-	if (create_server_record(&buf, ns->name, ns->id) < 0) {
+	record = server_record_new(ns->name, ns->id);
+	if (!record) {
 		error("Unable to allocate new service record");
-		dbus_message_unref(msg);
 		return 0;
 	}
 
-	dbus_message_append_args(msg, DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
-				&buf.data, buf.data_size, DBUS_TYPE_INVALID);
-
-	dbus_error_init(&derr);
-	reply = dbus_connection_send_with_reply_and_block(connection,
-			msg, -1, &derr);
-
-	free(buf.data);
-	dbus_message_unref(msg);
-
-	if (dbus_error_is_set(&derr) || dbus_set_error_from_message(&derr, reply)) {
-		error("Adding service record failed: %s", derr.message);
-		dbus_error_free(&derr);
+	if (add_record_to_server(&ns->src, record) < 0) {
+		error("Failed to register service record");
+		sdp_record_free(record);
 		return 0;
 	}
 
-	dbus_message_get_args(reply, &derr, DBUS_TYPE_UINT32, &rec_id,
-				DBUS_TYPE_INVALID);
+	debug("register_server_record: got record id 0x%x", record->handle);
 
-	if (dbus_error_is_set(&derr)) {
-		error("Invalid arguments to AddServiceRecord reply: %s", derr.message);
-		dbus_message_unref(reply);
-		dbus_error_free(&derr);
-		return 0;
-	}
-
-	dbus_message_unref(reply);
-
-	debug("add_server_record: got record id 0x%x", rec_id);
-
-	return rec_id;
+	return record->handle;
 }
 
-static int update_server_record(struct network_server *ns)
-{
-	DBusMessage *msg, *reply;
-	DBusError derr;
-	sdp_buf_t buf;
-
-	msg = dbus_message_new_method_call("org.bluez", "/org/bluez",
-				"org.bluez.Database", "UpdateServiceRecord");
-	if (!msg) {
-		error("Can't allocate new method call");
-		return -ENOMEM;
-	}
-
-	if (create_server_record(&buf, ns->name, ns->id) < 0) {
-		error("Unable to allocate new service record");
-		dbus_message_unref(msg);
-		return -1;
-	}
-
-	dbus_message_append_args(msg,
-			DBUS_TYPE_UINT32, &ns->record_id,
-			DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
-			&buf.data, buf.data_size, DBUS_TYPE_INVALID);
-
-	dbus_error_init(&derr);
-	reply = dbus_connection_send_with_reply_and_block(connection,
-			msg, -1, &derr);
-
-	free(buf.data);
-	dbus_message_unref(msg);
-
-	if (dbus_error_is_set(&derr) || dbus_set_error_from_message(&derr, reply)) {
-		error("Update service record failed: %s", derr.message);
-		dbus_error_free(&derr);
-		return -1;
-	}
-
-	dbus_message_unref(reply);
-
-	return 0;
-}
-
-static int remove_server_record(uint32_t rec_id)
-{
-	DBusMessage *msg, *reply;
-	DBusError derr;
-
-	msg = dbus_message_new_method_call("org.bluez", "/org/bluez",
-				"org.bluez.Database", "RemoveServiceRecord");
-	if (!msg) {
-		error("Can't allocate new method call");
-		return -ENOMEM;
-	}
-
-	dbus_message_append_args(msg,
-			DBUS_TYPE_UINT32, &rec_id,
-			DBUS_TYPE_INVALID);
-
-	dbus_error_init(&derr);
-	reply = dbus_connection_send_with_reply_and_block(connection,
-			msg, -1, &derr);
-
-	dbus_message_unref(msg);
-
-	if (dbus_error_is_set(&derr)) {
-		error("Removing service record 0x%x failed: %s",
-						rec_id, derr.message);
-		dbus_error_free(&derr);
-		return -1;
-	}
-
-	dbus_message_unref(reply);
-
-	return 0;
-}
-
-static DBusHandlerResult get_uuid(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+static DBusMessage *get_uuid(DBusConnection *conn,
+				DBusMessage *msg, void *data)
 {
 	struct network_server *ns = data;
 	DBusMessage *reply;
@@ -833,31 +645,46 @@ static DBusHandlerResult get_uuid(DBusConnection *conn,
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return NULL;
 
 	uuid = bnep_uuid(ns->id);
 	dbus_message_append_args(reply,
 			DBUS_TYPE_STRING, &uuid,
 			DBUS_TYPE_INVALID);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
-static DBusHandlerResult enable(DBusConnection *conn,
-				DBusMessage *msg, void *data)
+static inline DBusMessage *failed(DBusMessage *msg, const char *description)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".Failed",
+				description);
+}
+
+static inline DBusMessage *invalid_arguments(DBusMessage *msg,
+					const char *description)
+{
+	return g_dbus_create_error(msg, ERROR_INTERFACE ".InvalidArguments",
+				description);
+}
+
+static DBusMessage *enable(DBusConnection *conn,
+			DBusMessage *msg, void *data)
 {
 	struct network_server *ns = data;
 	DBusMessage *reply;
 
 	if (ns->enable)
-		return error_already_exists(conn, msg, "Server already enabled");
+		return g_dbus_create_error(msg, ERROR_INTERFACE
+						".AlreadyExist",
+						"Server already enabled");
 
 	if (bacmp(&ns->src, BDADDR_ANY) == 0) {
 		int dev_id;
 
 		dev_id = hci_get_route(&ns->src);
 		if ((dev_id < 0) || (hci_devba(dev_id, &ns->src) < 0))
-			return error_failed(conn, msg, "Adapter not available");
+			return failed(msg, "Adapter not available");
 
 		/* Store the server info */
 		server_store(ns->path);
@@ -865,24 +692,23 @@ static DBusHandlerResult enable(DBusConnection *conn,
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return NULL;
 
 	/* Add the service record */
-	ns->record_id = add_server_record(ns);
+	ns->record_id = register_server_record(ns);
 	if (!ns->record_id) {
 		dbus_message_unref(reply);
-		return error_failed(conn, msg,
-				"service record registration failed");
+		return failed(msg, "Service record registration failed");
 	}
 
 	ns->enable = TRUE;
 
 	store_property(&ns->src, ns->id, "enabled", "1");
 
-	dbus_connection_emit_signal(conn, ns->path, NETWORK_SERVER_INTERFACE,
+	g_dbus_emit_signal(conn, ns->path, NETWORK_SERVER_INTERFACE,
 					"Enabled", DBUS_TYPE_INVALID);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
 static void kill_connection(void *data, void *udata)
@@ -894,22 +720,22 @@ static void kill_connection(void *data, void *udata)
 	bnep_kill_connection(&dst);
 }
 
-static DBusHandlerResult disable(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+static DBusMessage *disable(DBusConnection *conn,
+				DBusMessage *msg, void *data)
 {
 	struct network_server *ns = data;
 	DBusMessage *reply;
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return NULL;
 
 	if (!ns->enable)
-		return error_failed(conn, msg, "Not enabled");
+		return failed(msg, "Not enabled");
 
 	/* Remove the service record */
 	if (ns->record_id) {
-		remove_server_record(ns->record_id);
+		remove_record_from_server(ns->record_id);
 		ns->record_id = 0;
 	}
 
@@ -919,71 +745,70 @@ static DBusHandlerResult disable(DBusConnection *conn,
 
 	store_property(&ns->src, ns->id, "enabled", "0");
 
-	dbus_connection_emit_signal(conn, ns->path, NETWORK_SERVER_INTERFACE,
+	g_dbus_emit_signal(conn, ns->path, NETWORK_SERVER_INTERFACE,
 					"Disabled", DBUS_TYPE_INVALID);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
-static DBusHandlerResult is_enabled(DBusConnection *conn, DBusMessage *msg,
-					void *data)
+static DBusMessage *is_enabled(DBusConnection *conn, DBusMessage *msg,
+				void *data)
 {
 	struct network_server *ns = data;
 	DBusMessage *reply;
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return NULL;
 
 	dbus_message_append_args(reply, DBUS_TYPE_BOOLEAN, &ns->enable,
 					DBUS_TYPE_INVALID);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
-static DBusHandlerResult set_name(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+static DBusMessage *set_name(DBusConnection *conn,
+				DBusMessage *msg, void *data)
 {
 	struct network_server *ns = data;
 	DBusMessage *reply;
-	DBusError derr;
 	const char *name;
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return NULL;
 
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
+	if (!dbus_message_get_args(msg, NULL,
 				DBUS_TYPE_STRING, &name,
-				DBUS_TYPE_INVALID)) {
-		error_invalid_arguments(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+				DBUS_TYPE_INVALID))
+		return NULL;
 
 	if (!name || (strlen(name) == 0))
-		return error_invalid_arguments(conn, msg, "Invalid name");
+		return invalid_arguments(msg, "Invalid name");
 
 	if (ns->name)
 		g_free(ns->name);
 	ns->name = g_strdup(name);
 
-	if (ns->enable) {
-		if (update_server_record(ns) < 0) {
+	if (ns->enable && ns->record_id) {
+		uint32_t handle = register_server_record(ns);
+		if (!handle) {
 			dbus_message_unref(reply);
-			return error_failed(conn, msg,
+			return failed(msg,
 				"Service record attribute update failed");
 		}
+
+		remove_record_from_server(ns->record_id);
+		ns->record_id = handle;
 	}
 
 	store_property(&ns->src, ns->id, "name", ns->name);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
-static DBusHandlerResult get_name(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+static DBusMessage *get_name(DBusConnection *conn,
+				DBusMessage *msg, void *data)
 {
 	struct network_server *ns = data;
 	char name[] = "";
@@ -992,55 +817,45 @@ static DBusHandlerResult get_name(DBusConnection *conn,
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return NULL;
 
 	dbus_message_append_args(reply,
 			DBUS_TYPE_STRING, &pname,
 			DBUS_TYPE_INVALID);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
-static DBusHandlerResult set_address_range(DBusConnection *conn,
+static DBusMessage *set_address_range(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
-	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+	return NULL;
 }
 
-static DBusHandlerResult set_routing(DBusConnection *conn, DBusMessage *msg,
+static DBusMessage *set_routing(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
 	struct network_server *ns = data;
-	DBusMessage *reply;
-	DBusError derr;
 	const char *iface;
 
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
-
-	dbus_error_init(&derr);
-	if (!dbus_message_get_args(msg, &derr,
+	if (!dbus_message_get_args(msg, NULL,
 				DBUS_TYPE_STRING, &iface,
-				DBUS_TYPE_INVALID)) {
-		error_invalid_arguments(conn, msg, derr.message);
-		dbus_error_free(&derr);
-		return DBUS_HANDLER_RESULT_HANDLED;
-	}
+				DBUS_TYPE_INVALID))
+		return NULL;
 
 	/* FIXME: Check if the interface is valid/UP */
 	if (!iface || (strlen(iface) == 0))
-		return error_invalid_arguments(conn, msg, "Invalid interface");
+		return invalid_arguments(msg, "Invalid interface");
 
 	if (ns->iface)
 		g_free(ns->iface);
 	ns->iface = g_strdup(iface);
 
-	return send_message_and_unref(conn, reply);
+	return dbus_message_new_method_return(msg);
 }
 
-static DBusHandlerResult get_info(DBusConnection *conn,
-					DBusMessage *msg, void *data)
+static DBusMessage *get_info(DBusConnection *conn,
+				DBusMessage *msg, void *data)
 {
 	struct network_server *ns = data;
 	DBusMessage *reply;
@@ -1050,7 +865,7 @@ static DBusHandlerResult get_info(DBusConnection *conn,
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
-		return DBUS_HANDLER_RESULT_NEED_MEMORY;
+		return NULL;
 
 	dbus_message_iter_init_append(reply, &iter);
 
@@ -1068,7 +883,7 @@ static DBusHandlerResult get_info(DBusConnection *conn,
 
 	dbus_message_iter_close_container(&iter, &dict);
 
-	return send_message_and_unref(conn, reply);
+	return reply;
 }
 
 static void server_free(struct network_server *ns)
@@ -1078,7 +893,7 @@ static void server_free(struct network_server *ns)
 
 	/* FIXME: Missing release/free all bnepX interfaces */
 	if (ns->record_id)
-		remove_server_record(ns->record_id);
+		remove_record_from_server(ns->record_id);
 
 	if (ns->iface)
 		g_free(ns->iface);
@@ -1100,32 +915,33 @@ static void server_free(struct network_server *ns)
 	g_free(ns);
 }
 
-static void server_unregister(DBusConnection *conn, void *data)
+static void server_unregister(void *data)
 {
 	struct network_server *ns = data;
 
 	info("Unregistered server path:%s", ns->path);
 
+	servers = g_slist_remove(servers, ns);
 	server_free(ns);
 }
 
-static DBusMethodVTable server_methods[] = {
-	{ "GetUUID",		get_uuid,		"",	"s"	},
-	{ "Enable",		enable,			"",	""	},
-	{ "Disable",		disable,		"",	""	},
-	{ "IsEnabled",		is_enabled,		"",	"b"	},
-	{ "SetName",		set_name,		"s",	""	},
-	{ "GetName",		get_name,		"",	"s"	},
-	{ "SetAddressRange",	set_address_range,	"ss",	""	},
-	{ "SetRouting",		set_routing,		"s",	""	},
-	{ "GetInfo",		get_info,		"",	"a{sv}"	},
-	{ NULL, NULL, NULL, NULL }
+static GDBusMethodTable server_methods[] = {
+	{ "GetUUID",		"",	"s",	get_uuid },
+	{ "Enable",		"",	"",	enable },
+	{ "Disable",		"",	"",	disable },
+	{ "IsEnabled",		"",	"b",	is_enabled },
+	{ "SetName",		"s",	"",	set_name },
+	{ "GetName",		"",	"s",	get_name },
+	{ "SetAddressRange",	"ss",	"",	set_address_range },
+	{ "SetRouting",		"s",	"",	set_routing },
+	{ "GetInfo",		"",	"a{sv}",get_info },
+	{ }
 };
 
-static DBusSignalVTable server_signals[] = {
+static GDBusSignalTable server_signals[] = {
 	{ "Enabled",	""	},
 	{ "Disabled",	""	},
-	{ NULL, NULL }
+	{ }
 };
 
 int server_register(const char *path, bdaddr_t *src, uint16_t id)
@@ -1137,20 +953,13 @@ int server_register(const char *path, bdaddr_t *src, uint16_t id)
 
 	ns = g_new0(struct network_server, 1);
 
-	if (!dbus_connection_create_object_path(connection, path, ns,
-						server_unregister)) {
-		error("D-Bus failed to register %s path", path);
-		server_free(ns);
-		return -1;
-	}
-
-	if (!dbus_connection_register_interface(connection, path,
-						NETWORK_SERVER_INTERFACE,
-						server_methods,
-						server_signals, NULL)) {
+	if (!g_dbus_register_interface(connection, path,
+					NETWORK_SERVER_INTERFACE,
+					server_methods, server_signals, NULL,
+					ns, server_unregister)) {
 		error("D-Bus failed to register %s interface",
 				NETWORK_SERVER_INTERFACE);
-		dbus_connection_destroy_object_path(connection, path);
+		server_free(ns);
 		return -1;
 	}
 
@@ -1166,6 +975,8 @@ int server_register(const char *path, bdaddr_t *src, uint16_t id)
 	ns->id = id;
 	bacpy(&ns->src, src);
 
+	servers = g_slist_append(servers, ns);
+
 	info("Registered server path:%s", path);
 
 	return 0;
@@ -1198,28 +1009,23 @@ int server_register_from_file(const char *path, const bdaddr_t *src,
 	str = textfile_get(filename, "enabled");
 	if (str) {
 		if (strcmp("1", str) == 0) {
-			ns->record_id = add_server_record(ns);
+			ns->record_id = register_server_record(ns);
 			ns->enable = TRUE;
 		}
 		g_free(str);
 	}
 
-	if (!dbus_connection_create_object_path(connection, path, ns,
-						server_unregister)) {
-		error("D-Bus failed to register %s path", path);
+	if (!g_dbus_register_interface(connection, path,
+					NETWORK_SERVER_INTERFACE,
+					server_methods, server_signals, NULL,
+					ns, server_unregister)) {
+		error("D-Bus failed to register %s interface",
+				NETWORK_SERVER_INTERFACE);
 		server_free(ns);
 		return -1;
 	}
 
-	if (!dbus_connection_register_interface(connection, path,
-						NETWORK_SERVER_INTERFACE,
-						server_methods,
-						server_signals, NULL)) {
-		error("D-Bus failed to register %s interface",
-				NETWORK_SERVER_INTERFACE);
-		dbus_connection_destroy_object_path(connection, path);
-		return -1;
-	}
+	servers = g_slist_append(servers, ns);
 
 	info("Registered server path:%s", path);
 
@@ -1231,13 +1037,15 @@ int server_store(const char *path)
 	struct network_server *ns;
 	char filename[PATH_MAX + 1];
 	char addr[18];
+	GSList *l;
 
-	if (!dbus_connection_get_object_user_data(connection,
-				path, (void *) &ns)) {
+	l = g_slist_find_custom(servers, path, find_server);
+	if (!l) {
 		error("Unable to salve %s on storage", path);
 		return -ENOENT;
 	}
 
+	ns = l->data;
 	ba2str(&ns->src, addr);
 	if (ns->id == BNEP_SVC_NAP)
 		create_name(filename, PATH_MAX, STORAGEDIR, addr, "nap");
@@ -1265,10 +1073,13 @@ int server_find_data(const char *path, const char *pattern)
 {
 	struct network_server *ns;
 	const char *uuid;
+	GSList *l;
 
-	if (!dbus_connection_get_object_user_data(connection, path, (void *) &ns))
+	l = g_slist_find_custom(servers, path, find_server);
+	if (!l)
 		return -1;
 
+	ns = l->data;
 	if (ns->name && strcasecmp(pattern, ns->name) == 0)
 		return 0;
 
