@@ -30,7 +30,7 @@
 
 #include <stdio.h>
 #include <errno.h>
-#include <malloc.h>
+#include <stdlib.h>
 #include <sys/time.h>
 #include <sys/socket.h>
 
