@@ -36,6 +36,7 @@
 #include <syslog.h>
 #include <signal.h>
 #include <sys/time.h>
+#include <sys/ioctl.h>
 #include <sys/socket.h>
 
 #include <bluetooth/bluetooth.h>
@@ -79,6 +80,7 @@ static int encrypt = 0;
 static int secure = 0;
 static int socktype = SOCK_STREAM;
 static int linger = 0;
+static int timestamp = 0;
 
 static float tv2fl(struct timeval tv)
 {
@@ -90,7 +92,7 @@ static int do_connect(char *svr)
 	struct sockaddr_rc addr;
 	struct rfcomm_conninfo conn;
 	socklen_t optlen;
-	int sk;
+	int sk, opt;
 
 	/* Create socket */
 	sk = socket(PF_BLUETOOTH, socktype, BTPROTO_RFCOMM);
@@ -100,6 +102,30 @@ static int do_connect(char *svr)
 		return -1;
 	}
 
+	/* Bind to local address */
+	memset(&addr, 0, sizeof(addr));
+	addr.rc_family = AF_BLUETOOTH;
+	bacpy(&addr.rc_bdaddr, &bdaddr);
+
+	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
+		syslog(LOG_ERR, "Can't bind socket: %s (%d)",
+							strerror(errno), errno);
+		goto error;
+	}
+
+#if 0
+	/* Enable SO_TIMESTAMP */
+	if (timestamp) {
+		int t = 1;
+
+		if (setsockopt(sk, SOL_SOCKET, SO_TIMESTAMP, &t, sizeof(t)) < 0) {
+			syslog(LOG_ERR, "Can't enable SO_TIMESTAMP: %s (%d)",
+							strerror(errno), errno);
+			goto error;
+		}
+	}
+#endif
+
 	/* Enable SO_LINGER */
 	if (linger) {
 		struct linger l = { .l_onoff = 1, .l_linger = linger };
@@ -111,13 +137,19 @@ static int do_connect(char *svr)
 		}
 	}
 
-	/* Bind to local address */
-	memset(&addr, 0, sizeof(addr));
-	addr.rc_family = AF_BLUETOOTH;
-	bacpy(&addr.rc_bdaddr, &bdaddr);
+	/* Set link mode */
+	opt = 0;
+	if (master)
+		opt |= RFCOMM_LM_MASTER;
+	if (auth)
+		opt |= RFCOMM_LM_AUTH;
+	if (encrypt)
+		opt |= RFCOMM_LM_ENCRYPT;
+	if (secure)
+		opt |= RFCOMM_LM_SECURE;
 
-	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
-		syslog(LOG_ERR, "Can't bind socket: %s (%d)",
+	if (opt && setsockopt(sk, SOL_RFCOMM, RFCOMM_LM, &opt, sizeof(opt)) < 0) {
+		syslog(LOG_ERR, "Can't set RFCOMM link mode: %s (%d)",
 							strerror(errno), errno);
 		goto error;
 	}
@@ -254,6 +286,19 @@ static void do_listen(void (*handler)(int sk))
 			ba, conn.hci_handle,
 			conn.dev_class[2], conn.dev_class[1], conn.dev_class[0]);
 
+#if 0
+		/* Enable SO_TIMESTAMP */
+		if (timestamp) {
+			int t = 1;
+
+			if (setsockopt(nsk, SOL_SOCKET, SO_TIMESTAMP, &t, sizeof(t)) < 0) {
+				syslog(LOG_ERR, "Can't enable SO_TIMESTAMP: %s (%d)",
+							strerror(errno), errno);
+				goto error;
+			}
+		}
+#endif
+
 		/* Enable SO_LINGER */
 		if (linger) {
 			struct linger l = { .l_onoff = 1, .l_linger = linger };
@@ -290,11 +335,14 @@ static void dump_mode(int sk)
 
 static void recv_mode(int sk)
 {
-	struct timeval tv_beg,tv_end,tv_diff;
+	struct timeval tv_beg, tv_end, tv_diff;
+	char ts[30];
 	long total;
 	uint32_t seq;
 
-	syslog(LOG_INFO,"Receiving ...");
+	syslog(LOG_INFO, "Receiving ...");
+
+	memset(ts, 0, sizeof(ts));
 
 	seq = 0;
 	while (1) {
@@ -311,6 +359,19 @@ static void recv_mode(int sk)
 							strerror(errno), errno);
 				return;	
 			}
+
+			if (timestamp) {
+				struct timeval tv;
+
+				if (ioctl(sk, SIOCGSTAMP, &tv) < 0) {
+					timestamp = 0;
+					memset(ts, 0, sizeof(ts));
+				} else {
+					sprintf(ts, "[%ld.%ld] ",
+							tv.tv_sec, tv.tv_usec);
+				}
+			}
+
 #if 0
 			/* Check sequence */
 			sq = btohl(*(uint32_t *) buf);
@@ -339,7 +400,7 @@ static void recv_mode(int sk)
 
 		timersub(&tv_end,&tv_beg,&tv_diff);
 
-		syslog(LOG_INFO,"%ld bytes in %.2f sec, %.2f kB/s", total,
+		syslog(LOG_INFO,"%s%ld bytes in %.2f sec, %.2f kB/s", ts, total,
 			tv2fl(tv_diff), (float)(total / tv2fl(tv_diff) ) / 1024.0);
 	}
 }
@@ -448,7 +509,8 @@ static void usage(void)
 		"\t[-A] request authentication\n"
 		"\t[-E] request encryption\n"
 		"\t[-S] secure connection\n"
-		"\t[-M] become master\n");
+		"\t[-M] become master\n"
+		"\t[-T] enable timestamps\n");
 }
 
 int main(int argc, char *argv[])
@@ -458,7 +520,7 @@ int main(int argc, char *argv[])
 
 	bacpy(&bdaddr, BDADDR_ANY);
 
-	while ((opt=getopt(argc,argv,"rdscuwmnb:i:P:B:N:MAESL:C:D:")) != EOF) {
+	while ((opt=getopt(argc,argv,"rdscuwmnb:i:P:B:N:MAESL:C:D:T")) != EOF) {
 		switch (opt) {
 		case 'r':
 			mode = RECV;
@@ -548,6 +610,10 @@ int main(int argc, char *argv[])
 			delay = atoi(optarg) * 1000;
 			break;
 
+		case 'T':
+			timestamp = 1;
+			break;
+
 		default:
 			usage();
 			exit(1);
